"""Tests related to uaclient.entitlement.__init__ module."""

import mock
import pytest

from uaclient import entitlements, exceptions, messages
from uaclient.entitlements.entitlement_status import ApplicabilityStatus
from uaclient.entitlements.repo import RepoEntitlement


class TestValidServices:
    @pytest.mark.parametrize("show_all_names", ((True), (False)))
    def test_valid_services(
        self,
        show_all_names,
        FakeConfig,
    ):
        m_cls_1 = mock.MagicMock()
        m_inst_1 = mock.MagicMock()
        type(m_cls_1).name = mock.PropertyMock(return_value="ent1")
        m_inst_1.presentation_name = "ent1"
        m_inst_1.valid_names = ["ent1", "othername"]
        m_cls_1.return_value = m_inst_1

        m_cls_2 = mock.MagicMock()
        m_inst_2 = mock.MagicMock()
        type(m_cls_2).name = mock.PropertyMock(return_value="ent2")
        m_inst_2.presentation_name = "ent2"
        m_inst_2.valid_names = ["ent2"]
        m_cls_2.return_value = m_inst_2

        ents = {m_cls_1, m_cls_2}

        with mock.patch.object(entitlements, "ENTITLEMENT_CLASSES", ents):
            expected_services = ["ent1", "ent2"]
            if show_all_names:
                expected_services.append("othername")

            assert expected_services == entitlements.valid_services(
                cfg=FakeConfig(), all_names=show_all_names
            )


class TestEntitlementFactory:
    def test_entitlement_factory(self, FakeConfig):
        m_cls_1 = mock.MagicMock()
        m_variant = mock.MagicMock()
        m_variant_obj = m_variant.return_value
        m_cls_1.return_value.valid_names = ["ent1", "othername"]
        m_cls_1.return_value.variants = {"variant1": m_variant}
        m_obj_1 = m_cls_1.return_value

        m_cls_2 = mock.MagicMock()
        m_cls_2.return_value.valid_names = ["ent2"]
        m_obj_2 = m_cls_2.return_value

        ents = {m_cls_1, m_cls_2}
        cfg = FakeConfig()

        with mock.patch.object(entitlements, "ENTITLEMENT_CLASSES", ents):
            assert m_obj_1 == entitlements.entitlement_factory(
                cfg=cfg, name="othername"
            )
            assert m_obj_2 == entitlements.entitlement_factory(
                cfg=cfg, name="ent2"
            )
            assert m_variant_obj == entitlements.entitlement_factory(
                cfg=cfg, name="ent1", variant="variant1"
            )
        with pytest.raises(exceptions.EntitlementNotFoundError):
            entitlements.entitlement_factory(cfg=cfg, name="nonexistent")

        with mock.patch.object(entitlements, "ENTITLEMENT_CLASSES", ents):
            with pytest.raises(exceptions.EntitlementNotFoundError) as excinfo:
                entitlements.entitlement_factory(
                    cfg=cfg, name="ent1", variant="nonexistent"
                )

        assert (
            messages.E_ENTITLEMENT_NOT_FOUND.format(
                entitlement_name="nonexistent"
            ).msg
            == excinfo.value.msg
        )


class TestSortEntitlements:
    def test_disable_order(self, FakeConfig):
        m_cls_1 = mock.MagicMock()
        m_obj_1 = m_cls_1.return_value
        m_obj_1.valid_names = ["ent1"]
        type(m_obj_1).dependent_services = mock.PropertyMock(return_value=())
        type(m_cls_1).name = mock.PropertyMock(return_value="ent1")

        m_cls_2 = mock.MagicMock()
        m_obj_2 = m_cls_2.return_value
        m_obj_2.valid_names = ["ent2"]
        type(m_obj_2).dependent_services = mock.PropertyMock(return_value=())
        type(m_cls_2).name = mock.PropertyMock(return_value="ent2")

        m_cls_3 = mock.MagicMock()
        m_obj_3 = m_cls_3.return_value
        m_obj_3.valid_names = ["ent3"]
        type(m_obj_3).dependent_services = mock.PropertyMock(
            return_value=(m_cls_1, m_cls_2)
        )
        type(m_cls_3).name = mock.PropertyMock(return_value="ent3")

        m_cls_5 = mock.MagicMock()
        m_obj_5 = m_cls_5.return_value
        m_obj_5.valid_names = ["ent5"]
        type(m_obj_5).dependent_services = mock.PropertyMock(return_value=())
        type(m_cls_5).name = mock.PropertyMock(return_value="ent5")

        m_cls_6 = mock.MagicMock()
        m_obj_6 = m_cls_6.return_value
        m_obj_6.valid_names = ["ent6"]
        type(m_obj_6).dependent_services = mock.PropertyMock(return_value=())
        type(m_cls_6).name = mock.PropertyMock(return_value="ent6")

        m_cls_4 = mock.MagicMock()
        m_obj_4 = m_cls_4.return_value
        m_obj_4.valid_names = ["ent4"]
        type(m_obj_4).dependent_services = mock.PropertyMock(
            return_value=(m_cls_5, m_cls_6)
        )
        type(m_cls_4).name = mock.PropertyMock(return_value="ent4")

        m_entitlements = [
            m_cls_1,
            m_cls_2,
            m_cls_3,
            m_cls_4,
            m_cls_5,
            m_cls_6,
        ]

        with mock.patch.object(
            entitlements, "ENTITLEMENT_CLASSES", m_entitlements
        ):
            assert [
                "ent1",
                "ent2",
                "ent3",
                "ent5",
                "ent6",
                "ent4",
            ] == entitlements.entitlements_disable_order(FakeConfig())

    def test_enable_order(self, FakeConfig):
        m_cls_2 = mock.MagicMock()
        m_obj_2 = m_cls_2.return_value
        m_obj_2.valid_names = ["ent2"]
        type(m_obj_2).required_services = mock.PropertyMock(return_value=())
        type(m_cls_2).name = mock.PropertyMock(return_value="ent2")

        m_cls_1 = mock.MagicMock()
        m_obj_1 = m_cls_1.return_value
        m_obj_1.valid_names = ["ent1"]
        type(m_obj_1).required_services = mock.PropertyMock(
            return_value=(mock.MagicMock(entitlement=m_cls_2),)
        )
        type(m_cls_1).name = mock.PropertyMock(return_value="ent1")

        m_cls_3 = mock.MagicMock()
        m_obj_3 = m_cls_3.return_value
        m_obj_3.valid_names = ["ent3"]
        type(m_obj_3).required_services = mock.PropertyMock(
            return_value=(
                mock.MagicMock(entitlement=m_cls_1),
                mock.MagicMock(entitlement=m_cls_2),
            )
        )
        type(m_cls_3).name = mock.PropertyMock(return_value="ent3")

        m_cls_5 = mock.MagicMock()
        m_obj_5 = m_cls_5.return_value
        m_obj_5.valid_names = ["ent5"]
        type(m_obj_5).required_services = mock.PropertyMock(return_value=())
        type(m_cls_5).name = mock.PropertyMock(return_value="ent5")

        m_cls_6 = mock.MagicMock()
        m_obj_6 = m_cls_6.return_value
        m_obj_6.valid_names = ["ent6"]
        type(m_obj_6).required_services = mock.PropertyMock(return_value=())
        type(m_cls_6).name = mock.PropertyMock(return_value="ent6")

        m_cls_4 = mock.MagicMock()
        m_obj_4 = m_cls_4.return_value
        m_obj_4.valid_names = ["ent4"]
        type(m_obj_4).required_services = mock.PropertyMock(
            return_value=(
                mock.MagicMock(entitlement=m_cls_5),
                mock.MagicMock(entitlement=m_cls_6),
            )
        )
        type(m_cls_4).name = mock.PropertyMock(return_value="ent4")

        m_entitlements = [
            m_cls_1,
            m_cls_2,
            m_cls_3,
            m_cls_4,
            m_cls_5,
            m_cls_6,
        ]

        with mock.patch.object(
            entitlements, "ENTITLEMENT_CLASSES", m_entitlements
        ):
            assert [
                "ent2",
                "ent1",
                "ent3",
                "ent5",
                "ent6",
                "ent4",
            ] == entitlements.entitlements_enable_order(cfg=FakeConfig())

    def test_order_entitlements_for_enabling(self, FakeConfig):
        m_cls_2 = mock.MagicMock()
        m_obj_2 = m_cls_2.return_value
        m_obj_2.valid_names = ["ent2"]
        type(m_obj_2).required_services = mock.PropertyMock(return_value=())
        type(m_cls_2).name = mock.PropertyMock(return_value="ent2")

        m_cls_1 = mock.MagicMock()
        m_obj_1 = m_cls_1.return_value
        m_obj_1.valid_names = ["ent1"]
        type(m_obj_1).required_services = mock.PropertyMock(
            return_value=(mock.MagicMock(entitlement=m_cls_2),)
        )
        type(m_cls_1).name = mock.PropertyMock(return_value="ent1")

        m_cls_3 = mock.MagicMock()
        m_obj_3 = m_cls_3.return_value
        m_obj_3.valid_names = ["ent3"]
        type(m_obj_3).required_services = mock.PropertyMock(
            return_value=(
                mock.MagicMock(entitlement=m_cls_1),
                mock.MagicMock(entitlement=m_cls_2),
            )
        )
        type(m_cls_3).name = mock.PropertyMock(return_value="ent3")

        m_cls_5 = mock.MagicMock()
        m_obj_5 = m_cls_5.return_value
        m_obj_5.valid_names = ["ent5"]
        type(m_obj_5).required_services = mock.PropertyMock(return_value=())
        type(m_cls_5).name = mock.PropertyMock(return_value="ent5")

        m_cls_6 = mock.MagicMock()
        m_obj_6 = m_cls_6.return_value
        m_obj_6.valid_names = ["ent6"]
        type(m_obj_6).required_services = mock.PropertyMock(return_value=())
        type(m_cls_6).name = mock.PropertyMock(return_value="ent6")

        m_cls_4 = mock.MagicMock()
        m_obj_4 = m_cls_4.return_value
        m_obj_4.valid_names = ["ent4"]
        type(m_obj_4).required_services = mock.PropertyMock(
            return_value=(
                mock.MagicMock(entitlement=m_cls_5),
                mock.MagicMock(entitlement=m_cls_6),
            )
        )
        type(m_cls_4).name = mock.PropertyMock(return_value="ent4")

        m_entitlements = [
            m_cls_1,
            m_cls_2,
            m_cls_3,
            m_cls_4,
            m_cls_5,
            m_cls_6,
        ]

        with mock.patch.object(
            entitlements, "ENTITLEMENT_CLASSES", m_entitlements
        ):
            assert [
                "ent2",
                "ent5",
                "ent4",
                "notthere",
                "ent6typo",
            ] == entitlements.order_entitlements_for_enabling(
                cfg=FakeConfig(),
                ents=["ent4", "notthere", "ent2", "ent6typo", "ent5"],
            )


class TestCheckEntitlementAPTDefinitionsAreUnique:
    @pytest.mark.parametrize(
        (
            "applicability_status1,applicability_status2,"
            "apt_url1,suite1,apt_url2,suite2,expected"
        ),
        (
            (
                (ApplicabilityStatus.APPLICABLE, None),
                (ApplicabilityStatus.APPLICABLE, None),
                "test",
                ("release",),
                "test",
                ("release",),
                exceptions.EntitlementsAPTDirectivesAreNotUnique(
                    url="test_url",
                    names="ent1, ent2",
                    apt_url="test",
                    suite="release",
                ),
            ),
            (
                (ApplicabilityStatus.APPLICABLE, None),
                (ApplicabilityStatus.INAPPLICABLE, None),
                "test",
                ("release",),
                "test",
                ("release",),
                True,
            ),
            (
                (ApplicabilityStatus.APPLICABLE, None),
                (ApplicabilityStatus.APPLICABLE, None),
                "test1",
                ("release1",),
                "test2",
                ("release2",),
                True,
            ),
            (
                (ApplicabilityStatus.APPLICABLE, None),
                (ApplicabilityStatus.APPLICABLE, None),
                "test1",
                ("release",),
                "test1",
                ("release2",),
                True,
            ),
            (
                (ApplicabilityStatus.APPLICABLE, None),
                (ApplicabilityStatus.APPLICABLE, None),
                "test1",
                ("release",),
                "test2",
                ("release",),
                True,
            ),
        ),
    )
    @mock.patch("uaclient.entitlements.entitlement_factory")
    @mock.patch("uaclient.entitlements.valid_services")
    def test_check_entitlement_definitions_are_unique(
        self,
        m_valid_services,
        m_ent_factory,
        applicability_status1,
        applicability_status2,
        apt_url1,
        suite1,
        apt_url2,
        suite2,
        expected,
    ):
        m_valid_services.return_value = ["ent1", "ent2"]

        m_ent1_obj = mock.MagicMock(RepoEntitlement)
        m_ent1_obj.applicability_status.return_value = applicability_status1
        type(m_ent1_obj).apt_url = apt_url1
        type(m_ent1_obj).apt_suites = suite1
        type(m_ent1_obj).repo_policy_check_tmpl = "{}/ubuntu {}"

        m_ent2_obj = mock.MagicMock(RepoEntitlement)
        m_ent2_obj.applicability_status.return_value = applicability_status2
        type(m_ent2_obj).apt_url = apt_url2
        type(m_ent2_obj).apt_suites = suite2
        type(m_ent2_obj).repo_policy_check_tmpl = "{}/ubuntu {}"

        m_ent_factory.side_effect = [m_ent1_obj, m_ent2_obj]

        if expected is True:
            assert entitlements.check_entitlement_apt_directives_are_unique(
                mock.MagicMock()
            )
        else:
            with pytest.raises(
                exceptions.EntitlementsAPTDirectivesAreNotUnique
            ):
                entitlements.check_entitlement_apt_directives_are_unique(
                    mock.MagicMock(url="test_url")
                )


class TestGetValidEntitlementNames:
    @mock.patch(
        "uaclient.entitlements.valid_services",
        return_value=["ent1", "ent2", "ent3"],
    )
    def test_get_valid_entitlements(self, _m_valid_services, FakeConfig):
        service = ["ent1", "ent3", "ent4"]
        expected_ents_found = ["ent1", "ent3"]
        expected_ents_not_found = ["ent4"]

        actual_ents_found, actual_ents_not_found = (
            entitlements.get_valid_entitlement_names(service, cfg=FakeConfig())
        )

        assert expected_ents_found == actual_ents_found
        assert expected_ents_not_found == actual_ents_not_found



import mock
import pytest

from uaclient.cli.security_status import security_status_command
from uaclient.util import DatetimeAwareJSONEncoder

M_PATH = "uaclient.cli.security_status."


@mock.patch(M_PATH + "security_status.security_status")
@mock.patch(M_PATH + "security_status.security_status_dict")
@mock.patch("uaclient.contract.get_available_resources")
class TestActionSecurityStatus:
    @pytest.mark.parametrize("output_format", ("json", "yaml", "text"))
    @mock.patch(M_PATH + "safe_dump")
    @mock.patch(M_PATH + "json.dumps")
    def test_action_security_status(
        self,
        m_dumps,
        m_safe_dump,
        _m_resources,
        m_security_status_dict,
        m_security_status,
        output_format,
        FakeConfig,
    ):
        cfg = FakeConfig()
        args = mock.MagicMock()
        args.format = output_format
        args.thirdparty = False
        args.unavailable = False
        args.esm_infra = False
        args.esm_apps = False
        security_status_command.action(args, cfg=cfg)

        if output_format == "json":
            assert m_dumps.call_args_list == [
                mock.call(
                    m_security_status_dict.return_value,
                    sort_keys=True,
                    cls=DatetimeAwareJSONEncoder,
                ),
            ]
            assert m_safe_dump.call_count == 0
        elif output_format == "yaml":
            assert m_safe_dump.call_args_list == [
                mock.call(
                    m_security_status_dict.return_value,
                    default_flow_style=False,
                )
            ]
            assert m_dumps.call_count == 0
        else:
            assert m_dumps.call_count == 0
            assert m_safe_dump.call_count == 0
            assert m_security_status.call_args_list == [mock.call(cfg)]
            assert m_security_status.call_count == 1



import datetime
import os
from typing import Dict, List, Optional, Tuple, Union

from uaclient import apt, exceptions, messages, system
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.exceptions import UnattendedUpgradesError
from uaclient.apt import get_apt_config_keys, get_apt_config_values
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    DatetimeDataValue,
    Field,
    IntDataValue,
    StringDataValue,
    data_list,
)

UNATTENDED_UPGRADES_CONFIG_KEYS = [
    "APT::Periodic::Enable",
    "APT::Periodic::Update-Package-Lists",
    "APT::Periodic::Unattended-Upgrade",
    "Unattended-Upgrade::Allowed-Origins",
]

UNATTENDED_UPGRADES_STAMP_PATH = "/var/lib/apt/periodic/upgrade-stamp"


class UnattendedUpgradesDisabledReason(DataObject):
    fields = [
        Field("msg", StringDataValue, doc="Human readable reason"),
        Field("code", StringDataValue, doc="Reason code"),
    ]

    def __init__(self, msg: str, code: str):
        self.msg = msg
        self.code = code


class UnattendedUpgradesStatusResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "systemd_apt_timer_enabled",
            BoolDataValue,
            doc="Indicate if the ``apt-daily.timer`` jobs are enabled",
        ),
        Field(
            "apt_periodic_job_enabled",
            BoolDataValue,
            doc=(
                "Indicate if the ``APT::Periodic::Enabled`` configuration is"
                " turned off"
            ),
        ),
        Field(
            "package_lists_refresh_frequency_days",
            IntDataValue,
            doc=(
                "The value of the ``APT::Periodic::Update-Package-Lists``"
                " configuration"
            ),
        ),
        Field(
            "unattended_upgrades_frequency_days",
            IntDataValue,
            doc=(
                "The value of the ``APT::Periodic::Unattended-Upgrade``"
                " configuration"
            ),
        ),
        Field(
            "unattended_upgrades_allowed_origins",
            data_list(StringDataValue),
            doc=(
                "The value of the ``Unattended-Upgrade::Allowed-Origins``"
                " configuration"
            ),
        ),
        Field(
            "unattended_upgrades_running",
            BoolDataValue,
            doc=(
                "Indicate if the ``unattended-upgrade`` service is correctly"
                " configured and running"
            ),
        ),
        Field(
            "unattended_upgrades_disabled_reason",
            UnattendedUpgradesDisabledReason,
            required=False,
            doc=(
                "Object that explains why ``unattended-upgrades`` is not"
                " running -- if the application is running, the object will be"
                " null"
            ),
        ),
        Field(
            "unattended_upgrades_last_run",
            DatetimeDataValue,
            required=False,
            doc="The last time ``unattended-upgrades`` ran",
        ),
    ]

    def __init__(
        self,
        *,
        systemd_apt_timer_enabled: bool,
        apt_periodic_job_enabled: bool,
        package_lists_refresh_frequency_days: int,
        unattended_upgrades_frequency_days: int,
        unattended_upgrades_allowed_origins: List[str],
        unattended_upgrades_running: bool,
        unattended_upgrades_disabled_reason: Optional[
            UnattendedUpgradesDisabledReason
        ],
        unattended_upgrades_last_run: Optional[datetime.datetime]
    ):
        self.systemd_apt_timer_enabled = systemd_apt_timer_enabled
        self.apt_periodic_job_enabled = apt_periodic_job_enabled
        self.package_lists_refresh_frequency_days = (
            package_lists_refresh_frequency_days
        )
        self.unattended_upgrades_frequency_days = (
            unattended_upgrades_frequency_days
        )
        self.unattended_upgrades_allowed_origins = (
            unattended_upgrades_allowed_origins
        )
        self.unattended_upgrades_running = unattended_upgrades_running
        self.unattended_upgrades_disabled_reason = (
            unattended_upgrades_disabled_reason
        )
        self.unattended_upgrades_last_run = unattended_upgrades_last_run


def _get_apt_daily_job_status() -> bool:
    try:
        apt_daily_job_enabled = system.is_systemd_unit_active(
            "apt-daily.timer"
        )
        apt_daily_upgrade_job_enabled = system.is_systemd_unit_active(
            "apt-daily-upgrade.timer"
        )
        systemd_apt_timer_enabled = (
            apt_daily_job_enabled and apt_daily_upgrade_job_enabled
        )
    except exceptions.ProcessExecutionError as e:
        raise UnattendedUpgradesError(error_msg=str(e))

    return systemd_apt_timer_enabled


def _is_unattended_upgrades_running(
    systemd_apt_timer_enabled: bool,
    unattended_upgrades_cfg: Dict[str, Union[str, List[str]]],
) -> Tuple[bool, Optional[messages.NamedMessage]]:
    if not systemd_apt_timer_enabled:
        return (False, messages.UNATTENDED_UPGRADES_SYSTEMD_JOB_DISABLED)

    for key in UNATTENDED_UPGRADES_CONFIG_KEYS:
        value = unattended_upgrades_cfg.get(key)
        if isinstance(value, list) and not value:
            return (
                False,
                messages.UNATTENDED_UPGRADES_CFG_LIST_VALUE_EMPTY.format(
                    cfg_name=key
                ),
            )
        if value == "0" or not value:
            return (
                False,
                messages.UNATTENDED_UPGRADES_CFG_VALUE_TURNED_OFF.format(
                    cfg_name=key
                ),
            )

    return (True, None)


def _get_unattended_upgrades_last_run() -> Optional[datetime.datetime]:
    try:
        creation_epoch = os.path.getctime(UNATTENDED_UPGRADES_STAMP_PATH)
    except FileNotFoundError:
        return None

    return datetime.datetime.fromtimestamp(creation_epoch)


def status() -> UnattendedUpgradesStatusResult:
    return _status(UAConfig())


def _status(cfg: UAConfig) -> UnattendedUpgradesStatusResult:
    """
    This endpoint returns the status around ``unattended-upgrades``. The focus
    of the endpoint is to verify if the application is running and how it is
    configured on the machine.

    .. important::

        For this endpoint, we deliver a unique key under ``meta`` called
        ``raw_config``. This field contains all related ``unattended-upgrades``
        configurations, unparsed. This means that this field will maintain both
        original name and values for those configurations.
    """
    if not apt.is_installed("unattended-upgrades"):
        return UnattendedUpgradesStatusResult(
            systemd_apt_timer_enabled=False,
            apt_periodic_job_enabled=False,
            package_lists_refresh_frequency_days=0,
            unattended_upgrades_frequency_days=0,
            unattended_upgrades_allowed_origins=[],
            unattended_upgrades_disabled_reason=UnattendedUpgradesDisabledReason(  # noqa
                msg=messages.UNATTENDED_UPGRADES_UNINSTALLED.msg,
                code=messages.UNATTENDED_UPGRADES_UNINSTALLED.name,
            ),
            unattended_upgrades_running=False,
            unattended_upgrades_last_run=None,
        )

    systemd_apt_timer_enabled = _get_apt_daily_job_status()
    unattended_upgrades_last_run = _get_unattended_upgrades_last_run()

    unattended_upgrades_cfg = get_apt_config_values(
        set(
            UNATTENDED_UPGRADES_CONFIG_KEYS
            + get_apt_config_keys("Unattended-Upgrade")
        )
    )

    # If that key is not present on the APT Config, we assume it
    # that the config is "enabled", as by default this configuration
    # will not be present in APT
    unattended_upgrades_cfg["APT::Periodic::Enable"] = (
        unattended_upgrades_cfg["APT::Periodic::Enable"] or "1"
    )

    # This should be a list, even if the key is missing.
    unattended_upgrades_cfg["Unattended-Upgrade::Allowed-Origins"] = (
        unattended_upgrades_cfg["Unattended-Upgrade::Allowed-Origins"] or []
    )

    (
        unattended_upgrades_running,
        disabled_reason,
    ) = _is_unattended_upgrades_running(
        systemd_apt_timer_enabled, unattended_upgrades_cfg
    )

    if disabled_reason:
        unattended_upgrades_disabled_reason = UnattendedUpgradesDisabledReason(
            msg=disabled_reason.msg,
            code=disabled_reason.name,
        )
    else:
        unattended_upgrades_disabled_reason = None

    unattended_upgrades_result = UnattendedUpgradesStatusResult(
        systemd_apt_timer_enabled=systemd_apt_timer_enabled,
        apt_periodic_job_enabled=str(
            unattended_upgrades_cfg.get("APT::Periodic::Enable", "")
        )
        == "1",
        package_lists_refresh_frequency_days=int(
            unattended_upgrades_cfg.get(  # type: ignore
                "APT::Periodic::Update-Package-Lists"
            )
            or 0
        ),
        unattended_upgrades_frequency_days=int(
            unattended_upgrades_cfg.get(  # type: ignore
                "APT::Periodic::Unattended-Upgrade"
            )
            or 0
        ),
        unattended_upgrades_allowed_origins=list(
            unattended_upgrades_cfg.get("Unattended-Upgrade::Allowed-Origins")
            or []
        ),
        unattended_upgrades_disabled_reason=(
            unattended_upgrades_disabled_reason
        ),
        unattended_upgrades_running=unattended_upgrades_running,
        unattended_upgrades_last_run=unattended_upgrades_last_run,
    )
    unattended_upgrades_result.meta = {"raw_config": unattended_upgrades_cfg}

    return unattended_upgrades_result


endpoint = APIEndpoint(
    version="v1",
    name="UnattendedUpgradesStatus",
    fn=_status,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.14",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.unattended_upgrades.status.v1 import status

result = status()
""",
    "result_class": UnattendedUpgradesStatusResult,
    "exceptions": [
        (
            UnattendedUpgradesError,
            "Raised if we cannot run a necessary command to show the status of ``unattended-upgrades``",  # noqa: E501
        )
    ],
    "example_cli": "pro api u.unattended_upgrades.status.v1",
    "example_json": """
{
    "apt_periodic_job_enabled": true,
    "package_lists_refresh_frequency_days": 1,
    "systemd_apt_timer_enabled": true,
    "unattended_upgrades_allowed_origins": [
        "${distro_id}:${distro_codename}",
        "${distro_id}:${distro_codename}-security",
        "${distro_id}ESMApps:${distro_codename}-apps-security",
        "${distro_id}ESM:${distro_codename}-infra-security"
    ],
    "unattended_upgrades_disabled_reason": null,
    "unattended_upgrades_frequency_days": 1,
    "unattended_upgrades_last_run": null,
    "unattended_upgrades_running": true
}
""",
    "extra": """
- Possible attributes in JSON ``meta`` field:

  .. code-block:: json

     {
         "meta": {
             "environment_vars": [],
             "raw_config": {
                 "APT::Periodic::Enable": "1",
                 "APT::Periodic::Unattended-Upgrade": "1",
                 "APT::Periodic::Update-Package-Lists": "1",
                 "Unattended-Upgrade::Allowed-Origins": [
                     "${distro_id}:${distro_codename}",
                     "${distro_id}:${distro_codename}-security",
                     "${distro_id}ESMApps:${distro_codename}-apps-security",
                     "${distro_id}ESM:${distro_codename}-infra-security"
                 ]
             }
         }
     }
""",
    "extra_indent": 2,
}



import mock
import pytest

from uaclient import livepatch
from uaclient.api.u.pro.security.status.reboot_required.v1 import (
    RebootStatus,
    _get_reboot_status,
    _reboot_required,
)
from uaclient.system import RebootRequiredPkgs

M_PATH = "uaclient.api.u.pro.security.status.reboot_required.v1."


class TestRebootRequired:
    @pytest.mark.parametrize(
        (
            "reboot_state,reboot_required_pkgs,livepatch_status,"
            "expected_standard_pkgs,expected_kernel_pkgs,"
            "expected_livepatch_enabled_and_kernel,"
            "expected_livepatch_enabled,expected_livepatch_state,"
            "expected_livepatch_support"
        ),
        (
            (
                RebootStatus.REBOOT_REQUIRED,
                RebootRequiredPkgs(
                    standard_packages=["pkg1"], kernel_packages=["linux-base"]
                ),
                livepatch.LivepatchStatusStatus(
                    kernel=None,
                    livepatch=livepatch.LivepatchPatchStatus(
                        state="nothing-to-apply", fixes=None, version=None
                    ),
                    supported="supported",
                ),
                ["pkg1"],
                ["linux-base"],
                True,
                True,
                "nothing-to-apply",
                "supported",
            ),
            (
                RebootStatus.REBOOT_NOT_REQUIRED,
                None,
                None,
                None,
                None,
                False,
                False,
                None,
                None,
            ),
            (
                RebootStatus.REBOOT_REQUIRED_LIVEPATCH_APPLIED,
                RebootRequiredPkgs(
                    standard_packages=[], kernel_packages=["linux-base"]
                ),
                livepatch.LivepatchStatusStatus(
                    kernel=None,
                    livepatch=livepatch.LivepatchPatchStatus(
                        state="applied", fixes=None, version=None
                    ),
                    supported="supported",
                ),
                [],
                ["linux-base"],
                True,
                True,
                "applied",
                "supported",
            ),
            (
                RebootStatus.REBOOT_REQUIRED_LIVEPATCH_APPLIED,
                RebootRequiredPkgs(
                    standard_packages=[], kernel_packages=["linux-base"]
                ),
                livepatch.LivepatchStatusStatus(
                    kernel=None,
                    livepatch=livepatch.LivepatchPatchStatus(
                        state="unknown", fixes=None, version=None
                    ),
                    supported="unknown",
                ),
                [],
                ["linux-base"],
                False,
                True,
                "unknown",
                "unknown",
            ),
        ),
    )
    @mock.patch("uaclient.livepatch.status")
    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "_get_reboot_status")
    def test_reboot_status_api(
        self,
        m_get_reboot_status,
        m_get_reboot_required_pkgs,
        m_livepatch_status,
        reboot_state,
        reboot_required_pkgs,
        livepatch_status,
        expected_standard_pkgs,
        expected_kernel_pkgs,
        expected_livepatch_enabled_and_kernel,
        expected_livepatch_enabled,
        expected_livepatch_state,
        expected_livepatch_support,
    ):
        m_get_reboot_status.return_value = reboot_state
        m_get_reboot_required_pkgs.return_value = reboot_required_pkgs
        m_livepatch_status.return_value = livepatch_status
        result = _reboot_required(mock.MagicMock())
        assert reboot_state.value == result.reboot_required
        assert (
            expected_standard_pkgs
            == result.reboot_required_packages.standard_packages
        )
        assert (
            expected_livepatch_enabled_and_kernel
            == result.livepatch_enabled_and_kernel_patched
        )
        assert expected_livepatch_enabled == result.livepatch_enabled
        assert expected_livepatch_state == result.livepatch_state
        assert expected_livepatch_support == result.livepatch_support


class TestGetRebootStatus:
    @mock.patch(M_PATH + "should_reboot", return_value=False)
    def test_get_reboot_status_no_reboot_needed(self, m_should_reboot):
        assert _get_reboot_status() == RebootStatus.REBOOT_NOT_REQUIRED
        assert 1 == m_should_reboot.call_count

    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "should_reboot", return_value=True)
    def test_get_reboot_status_no_reboot_pkgs_file(
        self, m_should_reboot, m_get_reboot_required_pkgs
    ):
        m_get_reboot_required_pkgs.return_value = None
        assert _get_reboot_status() == RebootStatus.REBOOT_REQUIRED
        assert 1 == m_should_reboot.call_count
        assert 1 == m_get_reboot_required_pkgs.call_count

    @mock.patch("uaclient.livepatch.status")
    @mock.patch("uaclient.livepatch.is_livepatch_installed", return_value=True)
    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "should_reboot", return_value=True)
    def test_get_reboot_status_livepatch_status_none(
        self,
        m_should_reboot,
        m_get_reboot_required_pkgs,
        _m_is_livepatch_installed,
        m_livepatch_status,
    ):
        m_get_reboot_required_pkgs.return_value = RebootRequiredPkgs(
            standard_packages=[],
            kernel_packages=["linux-base", "linux-image-5.4.0-1074"],
        )
        m_livepatch_status.return_value = None
        assert _get_reboot_status() == RebootStatus.REBOOT_REQUIRED

    @pytest.mark.parametrize(
        "reboot_required_pkgs,expected_state",
        (
            (
                RebootRequiredPkgs(
                    standard_packages=["pkg1", "pkg2"], kernel_packages=[]
                ),
                RebootStatus.REBOOT_REQUIRED,
            ),
            (
                RebootRequiredPkgs(
                    standard_packages=["pkg2"],
                    kernel_packages=["linux-base", "linux-image-5.4.0-1074"],
                ),
                RebootStatus.REBOOT_REQUIRED,
            ),
        ),
    )
    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "should_reboot", return_value=True)
    def test_get_reboot_status_reboot_pkgs_file_present(
        self,
        m_should_reboot,
        m_get_reboot_required_pkgs,
        reboot_required_pkgs,
        expected_state,
    ):
        m_get_reboot_required_pkgs.return_value = reboot_required_pkgs
        assert _get_reboot_status() == expected_state
        assert 1 == m_should_reboot.call_count
        assert 1 == m_get_reboot_required_pkgs.call_count

    @pytest.mark.parametrize(
        [
            "livepatch_state",
            "supported_state",
            "expected_state",
            "kernel_name",
        ],
        (
            (
                "applied",
                "supported",
                RebootStatus.REBOOT_REQUIRED_LIVEPATCH_APPLIED,
                "4.15.0-187.198-generic",
            ),
            (
                "applied",
                None,
                RebootStatus.REBOOT_REQUIRED,
                "4.15.0-187.198-generic",
            ),
            ("applied", "supported", RebootStatus.REBOOT_REQUIRED, "test"),
            (
                "nothing-to-apply",
                "supported",
                RebootStatus.REBOOT_REQUIRED_LIVEPATCH_APPLIED,
                "4.15.0-187.198-generic",
            ),
            (
                "applying",
                "supported",
                RebootStatus.REBOOT_REQUIRED,
                "4.15.0-187.198-generic",
            ),
            (
                "apply-failed",
                "supported",
                RebootStatus.REBOOT_REQUIRED,
                "4.15.0-187.198-generic",
            ),
        ),
    )
    @mock.patch(M_PATH + "get_kernel_info")
    @mock.patch("uaclient.livepatch.is_livepatch_installed")
    @mock.patch("uaclient.livepatch.status")
    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "should_reboot", return_value=True)
    def test_get_reboot_status_reboot_pkgs_file_only_kernel_pkgs(
        self,
        m_should_reboot,
        m_get_reboot_required_pkgs,
        m_livepatch_status,
        m_is_livepatch_installed,
        m_kernel_info,
        livepatch_state,
        supported_state,
        expected_state,
        kernel_name,
    ):
        m_kernel_info.return_value = mock.MagicMock(
            proc_version_signature_version=kernel_name
        )
        m_is_livepatch_installed.return_value = True
        m_get_reboot_required_pkgs.return_value = RebootRequiredPkgs(
            standard_packages=[],
            kernel_packages=["linux-base", "linux-image-5.4.0-1074"],
        )
        m_livepatch_status.return_value = livepatch.LivepatchStatusStatus(
            kernel="4.15.0-187.198-generic",
            livepatch=livepatch.LivepatchPatchStatus(
                state=livepatch_state, fixes=None, version=None
            ),
            supported=supported_state,
        )

        assert _get_reboot_status() == expected_state
        assert 1 == m_should_reboot.call_count
        assert 1 == m_get_reboot_required_pkgs.call_count
        assert 1 == m_is_livepatch_installed.call_count
        assert 1 == m_livepatch_status.call_count
        assert 1 == m_kernel_info.call_count

    @mock.patch(M_PATH + "get_kernel_info")
    @mock.patch("uaclient.livepatch.is_livepatch_installed")
    @mock.patch("uaclient.livepatch.status")
    @mock.patch(M_PATH + "get_reboot_required_pkgs")
    @mock.patch(M_PATH + "should_reboot", return_value=True)
    def test_get_reboot_status_fail_parsing_kernel_info(
        self,
        m_should_reboot,
        m_get_reboot_required_pkgs,
        m_livepatch_status,
        m_is_livepatch_installed,
        m_kernel_info,
    ):
        m_kernel_info.return_value = mock.MagicMock(
            proc_version_signature_version=None
        )
        m_is_livepatch_installed.return_value = True
        m_get_reboot_required_pkgs.return_value = RebootRequiredPkgs(
            standard_packages=[],
            kernel_packages=["linux-base", "linux-image-5.4.0-1074"],
        )
        m_livepatch_status.return_value = livepatch.LivepatchStatusStatus(
            kernel="4.15.0-187.198-generic",
            livepatch=livepatch.LivepatchPatchStatus(
                state="applied", fixes=None, version=None
            ),
            supported=None,
        )

        assert _get_reboot_status() == RebootStatus.REBOOT_REQUIRED
        assert 1 == m_should_reboot.call_count
        assert 1 == m_get_reboot_required_pkgs.call_count
        assert 1 == m_is_livepatch_installed.call_count
        assert 1 == m_livepatch_status.call_count
        assert 1 == m_kernel_info.call_count



import mock
import pytest

from lib.add_esm_snapshot_auth import add_esm_snapshot_auth
from uaclient.entitlements.entitlement_status import ApplicationStatus

MOCK_ESM_INFRA_AUTH_UPDATE_CALL = mock.call(
    override_snapshot_urls=[
        "snapshot.infra-security.esm.ubuntu.com/infra/ubuntu/",
        "snapshot.infra-updates.esm.ubuntu.com/infra/ubuntu/",
    ]
)
MOCK_ESM_APPS_AUTH_UPDATE_CALL = mock.call(
    override_snapshot_urls=[
        "snapshot.apps-security.esm.ubuntu.com/apps/ubuntu/",
        "snapshot.apps-updates.esm.ubuntu.com/apps/ubuntu/",
    ]
)


class TestAddEsmSnapshotAuth:
    @pytest.mark.parametrize(
        [
            "is_attached",
            "esm_infra_application_status",
            "esm_apps_application_status",
            "esm_infra_apt_url",
            "esm_apps_apt_url",
            "esm_infra_update_apt_auth_calls",
            "esm_apps_update_apt_auth_calls",
        ],
        [
            (False, None, None, None, None, [], []),
            (
                True,
                ApplicationStatus.DISABLED,
                ApplicationStatus.DISABLED,
                None,
                None,
                [],
                [],
            ),
            (
                True,
                ApplicationStatus.ENABLED,
                ApplicationStatus.ENABLED,
                None,
                None,
                [],
                [],
            ),
            (
                True,
                ApplicationStatus.ENABLED,
                ApplicationStatus.ENABLED,
                "http://custom",
                "http://custom",
                [],
                [],
            ),
            (
                True,
                ApplicationStatus.ENABLED,
                ApplicationStatus.WARNING,
                "https://esm.ubuntu.com/infra",
                "https://esm.ubuntu.com/apps",
                [MOCK_ESM_INFRA_AUTH_UPDATE_CALL],
                [MOCK_ESM_APPS_AUTH_UPDATE_CALL],
            ),
            (
                True,
                ApplicationStatus.WARNING,
                ApplicationStatus.DISABLED,
                "https://esm.ubuntu.com/infra",
                "https://esm.ubuntu.com/apps",
                [MOCK_ESM_INFRA_AUTH_UPDATE_CALL],
                [],
            ),
            (
                True,
                ApplicationStatus.DISABLED,
                ApplicationStatus.ENABLED,
                "https://esm.ubuntu.com/infra",
                "https://esm.ubuntu.com/apps",
                [],
                [MOCK_ESM_APPS_AUTH_UPDATE_CALL],
            ),
        ],
    )
    @mock.patch("lib.add_esm_snapshot_auth.ESMAppsEntitlement.update_apt_auth")
    @mock.patch(
        "lib.add_esm_snapshot_auth.ESMInfraEntitlement.update_apt_auth"
    )
    @mock.patch(
        "lib.add_esm_snapshot_auth.ESMAppsEntitlement.apt_url",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "lib.add_esm_snapshot_auth.ESMInfraEntitlement.apt_url",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "lib.add_esm_snapshot_auth.ESMAppsEntitlement.application_status"
    )
    @mock.patch(
        "lib.add_esm_snapshot_auth.ESMInfraEntitlement.application_status"
    )
    @mock.patch("lib.add_esm_snapshot_auth._is_attached")
    def test_add_esm_snapshot_auth(
        self,
        m_is_attached,
        m_esm_infra_application_status,
        m_esm_apps_application_status,
        m_esm_infra_apt_url,
        m_esm_apps_apt_url,
        m_esm_infra_update_apt_auth,
        m_esm_apps_update_apt_auth,
        is_attached,
        esm_infra_application_status,
        esm_apps_application_status,
        esm_infra_apt_url,
        esm_apps_apt_url,
        esm_infra_update_apt_auth_calls,
        esm_apps_update_apt_auth_calls,
        FakeConfig,
    ):
        m_is_attached.return_value.is_attached = is_attached
        m_esm_infra_application_status.return_value = (
            esm_infra_application_status,
            None,
        )
        m_esm_apps_application_status.return_value = (
            esm_apps_application_status,
            None,
        )
        m_esm_infra_apt_url.return_value = esm_infra_apt_url
        m_esm_apps_apt_url.return_value = esm_apps_apt_url
        add_esm_snapshot_auth(FakeConfig())
        assert (
            m_esm_infra_update_apt_auth.call_args_list
            == esm_infra_update_apt_auth_calls
        )
        assert (
            m_esm_apps_update_apt_auth.call_args_list
            == esm_apps_update_apt_auth_calls
        )



import contextlib
import io
import json

import mock
import pytest

from uaclient.event_logger import JSON_SCHEMA_VERSION, EventLoggerMode
from uaclient.yaml import safe_load


class TestEventLogger:
    @pytest.mark.parametrize(
        "event_mode",
        (EventLoggerMode.CLI, EventLoggerMode.JSON, EventLoggerMode.YAML),
    )
    def test_process_events(self, event_mode, event):
        with mock.patch.object(event, "_event_logger_mode", event_mode):
            fake_stdout = io.StringIO()
            with contextlib.redirect_stdout(fake_stdout):
                event.info(info_msg="test")
                event.needs_reboot(reboot_required=True)
                event.service_processed("test")
                event.services_failed(["esm"])
                event.error(error_msg="error1", error_code="error1-code")
                event.error(error_msg="error2", service="esm")
                event.error(error_msg="error3", error_type="exception")
                event.error(error_msg="error4", additional_info={"test": 123})
                event.warning(warning_msg="warning1")
                event.warning(warning_msg="warning2", service="esm")
                event.process_events()

            expected_cli_out = "test"
            expected_machine_out = {
                "_schema_version": JSON_SCHEMA_VERSION,
                "result": "failure",
                "errors": [
                    {
                        "message": "error1",
                        "message_code": "error1-code",
                        "service": None,
                        "type": "system",
                    },
                    {
                        "message": "error2",
                        "message_code": None,
                        "service": "esm",
                        "type": "service",
                    },
                    {
                        "message": "error3",
                        "message_code": None,
                        "service": None,
                        "type": "exception",
                    },
                    {
                        "message": "error4",
                        "message_code": None,
                        "service": None,
                        "type": "system",
                        "additional_info": {"test": 123},
                    },
                ],
                "warnings": [
                    {
                        "message": "warning1",
                        "message_code": None,
                        "service": None,
                        "type": "system",
                    },
                    {
                        "message": "warning2",
                        "message_code": None,
                        "service": "esm",
                        "type": "service",
                    },
                ],
                "failed_services": ["esm"],
                "needs_reboot": True,
                "processed_services": ["test"],
            }

        if event_mode == EventLoggerMode.CLI:
            assert expected_cli_out == fake_stdout.getvalue().strip()
        elif event_mode == EventLoggerMode.JSON:
            assert expected_machine_out == json.loads(
                fake_stdout.getvalue().strip()
            )
        else:
            assert expected_machine_out == safe_load(
                fake_stdout.getvalue().strip()
            )

    @pytest.mark.parametrize(
        "env_return,env_list",
        (
            ({}, []),
            (
                {"UA_EXAMPLE_KEY1": "value1", "UA_EXAMPLE_KEY2": "value2"},
                [
                    {"name": "UA_EXAMPLE_KEY1", "value": "value1"},
                    {"name": "UA_EXAMPLE_KEY2", "value": "value2"},
                ],
            ),
        ),
    )
    @pytest.mark.parametrize(
        "event_mode",
        (EventLoggerMode.CLI, EventLoggerMode.JSON, EventLoggerMode.YAML),
    )
    @mock.patch("uaclient.util.get_pro_environment")
    def test_process_events_for_status(
        self, m_environment, event_mode, env_return, env_list, event
    ):
        m_environment.return_value = env_return
        with mock.patch.object(event, "_event_logger_mode", event_mode):
            fake_stdout = io.StringIO()
            with contextlib.redirect_stdout(fake_stdout):
                event.set_command("status")
                event.set_output_content(
                    {
                        "some_status_key": "some_status_information",
                        "a_list_of_things": ["first", "second", "third"],
                        "services": [],
                    }
                )
                event.info(info_msg="test")
                event.error(error_msg="error1")
                event.warning(warning_msg="warning1")
                event.process_events()

            expected_machine_out = {
                "some_status_key": "some_status_information",
                "a_list_of_things": ["first", "second", "third"],
                "environment_vars": env_list,
                "services": [],
                "result": "failure",
                "errors": [
                    {
                        "message": "error1",
                        "message_code": None,
                        "service": None,
                        "type": "system",
                    }
                ],
                "warnings": [
                    {
                        "message": "warning1",
                        "message_code": None,
                        "service": None,
                        "type": "system",
                    }
                ],
            }

            expected_cli_out = "test"

        if event_mode == EventLoggerMode.CLI:
            assert expected_cli_out == fake_stdout.getvalue().strip()
        elif event_mode == EventLoggerMode.JSON:
            assert expected_machine_out == json.loads(
                fake_stdout.getvalue().strip()
            )
        else:
            assert expected_machine_out == safe_load(
                fake_stdout.getvalue().strip()
            )



from uaclient import messages
from uaclient.api.data_types import APIResponse, ErrorWarningObject
from uaclient.exceptions import UbuntuProError
from uaclient.util import get_pro_environment
from uaclient.version import check_for_new_version


def error_out(exception: Exception) -> APIResponse:
    if isinstance(exception, (UbuntuProError, APIError)):
        error = ErrorWarningObject(
            title=exception.msg,
            code=exception.msg_code
            or "generic-" + exception.__class__.__name__,
            meta=exception.additional_info or {},
        )
    else:
        error = ErrorWarningObject(
            title=str(exception),
            code="generic-" + exception.__class__.__name__,
            meta={},
        )

    warnings = []
    new_version = check_for_new_version()
    if new_version:
        warnings.append(
            ErrorWarningObject(
                title=messages.WARN_NEW_VERSION_AVAILABLE.format(
                    version=new_version
                ).msg,
                code=messages.WARN_NEW_VERSION_AVAILABLE.name,
                meta={},
            )
        )

    return APIResponse(
        _schema_version="v1",
        result="failure",
        data={
            "meta": {
                "environment_vars": [
                    {"name": name, "value": value}
                    for name, value in sorted(get_pro_environment().items())
                ],
            }
        },
        errors=[error],
        warnings=warnings,
    )


class APIError(UbuntuProError):
    pass


class APIInvalidEndpoint(APIError):
    _formatted_msg = messages.E_API_INVALID_ENDPOINT


class APIMissingArg(APIError):
    _formatted_msg = messages.E_API_MISSING_ARG


class APINoArgsForEndpoint(APIError):
    _formatted_msg = messages.E_API_NO_ARG_FOR_ENDPOINT


class APIJSONDataFormatError(APIError):
    _formatted_msg = messages.E_API_JSON_DATA_FORMAT_ERROR


class APIBadArgsFormat(APIError):
    _formatted_msg = messages.E_API_BAD_ARGS_FORMAT



import mock
import pytest

from uaclient.api import exceptions
from uaclient.api.u.pro.services.disable.v1 import (
    DisableOptions,
    DisableResult,
    _disable,
)
from uaclient.testing.helpers import does_not_raise

M_PATH = "uaclient.api.u.pro.services.disable.v1."


class TestDisable:
    @pytest.mark.parametrize(
        [
            "options",
            "we_are_currently_root",
            "is_attached",
            "enabled_services_names_before",
            "enabled_services_names_after",
            "disable_result",
            "expected_raises",
            "expected_result",
        ],
        [
            # not root
            (
                DisableOptions(service="s1"),
                False,
                False,
                None,
                None,
                None,
                pytest.raises(exceptions.NonRootUserError),
                None,
            ),
            # not attached
            (
                DisableOptions(service="s1"),
                True,
                False,
                None,
                None,
                None,
                pytest.raises(exceptions.UnattachedError),
                None,
            ),
            # generic disable failure
            (
                DisableOptions(service="s1"),
                True,
                True,
                ["s1"],
                None,
                (False, None),
                pytest.raises(exceptions.EntitlementNotDisabledError),
                None,
            ),
            # success
            (
                DisableOptions(service="s1"),
                True,
                True,
                ["s1"],
                [],
                (True, None),
                does_not_raise(),
                DisableResult(
                    disabled=["s1"],
                ),
            ),
            # success already disabled
            (
                DisableOptions(service="s1"),
                True,
                True,
                [],
                None,
                None,
                does_not_raise(),
                DisableResult(
                    disabled=[],
                ),
            ),
            # success with additional disablements
            (
                DisableOptions(service="s1"),
                True,
                True,
                ["s1", "s2", "s3"],
                ["s2"],
                (True, None),
                does_not_raise(),
                DisableResult(
                    disabled=["s1", "s3"],
                ),
            ),
        ],
    )
    @mock.patch(M_PATH + "status.status")
    @mock.patch(M_PATH + "lock.clear_lock_file_if_present")
    @mock.patch(M_PATH + "lock.RetryLock")
    @mock.patch(M_PATH + "entitlements.entitlement_factory")
    @mock.patch(M_PATH + "_enabled_services_names")
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "util.we_are_currently_root")
    def test_disable(
        self,
        m_we_are_currently_root,
        m_is_attached,
        m_enabled_services_names,
        m_entitlement_factory,
        m_spin_lock,
        m_clear_lock_file_if_present,
        m_status,
        options,
        we_are_currently_root,
        is_attached,
        enabled_services_names_before,
        enabled_services_names_after,
        disable_result,
        expected_raises,
        expected_result,
        FakeConfig,
    ):
        m_we_are_currently_root.return_value = we_are_currently_root
        m_is_attached.return_value = mock.MagicMock(is_attached=is_attached)
        m_enabled_services_names.side_effect = [
            enabled_services_names_before,
            enabled_services_names_after,
        ]
        m_ent = m_entitlement_factory.return_value
        m_ent_variant = m_ent.enabled_variant
        m_ent_variant.disable.return_value = disable_result

        cfg = FakeConfig()

        actual_result = None
        with expected_raises:
            actual_result = _disable(
                options, cfg, progress_object=mock.MagicMock()
            )

        assert actual_result == expected_result

        if expected_result is not None and len(expected_result.disabled) > 0:
            assert m_entitlement_factory.call_args_list == [
                mock.call(
                    cfg=cfg,
                    name=options.service,
                    purge=options.purge,
                )
            ]
            assert m_ent_variant.disable.call_args_list == [
                mock.call(mock.ANY)
            ]



"""
Update messaging text for use in MOTD and APT custom Ubuntu Pro messages.

Messaging files will be emitted to /var/lib/ubuntu-advantage/message-* which
will be sourced by apt-hook/hook.cc and various /etc/update-motd.d/ hooks to
present updated text about Ubuntu Pro service and token state.
"""

import logging
import os
from os.path import exists

from uaclient import contract, defaults, messages, system, util
from uaclient.api.u.pro.packages.updates.v1 import (
    _updates as api_u_pro_packages_updates_v1,
)
from uaclient.api.u.pro.status.enabled_services.v1 import _enabled_services
from uaclient.api.u.pro.status.is_attached.v1 import (
    ContractExpiryStatus,
    _is_attached,
)
from uaclient.config import UAConfig
from uaclient.files import machine_token, notices

MOTD_CONTRACT_STATUS_FILE_NAME = "motd-contract-status"
UPDATE_NOTIFIER_MOTD_SCRIPT = (
    "/usr/lib/update-notifier/update-motd-updates-available"
)
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def update_contract_expiry(cfg: UAConfig):
    machine_token_file = machine_token.get_machine_token_file(cfg)
    orig_token = machine_token_file.machine_token
    contract_id = (
        orig_token.get("machineTokenInfo", {})
        .get("contractInfo", {})
        .get("id", None)
    )
    contract_client = contract.UAContractClient(cfg)
    resp = contract_client.get_contract_machine(
        orig_token.get("machineToken", ""), contract_id
    )
    resp_expiry = (
        resp.get("machineTokenInfo", {})
        .get("contractInfo", {})
        .get("effectiveTo", None)
    )
    if (
        resp_expiry is not None
        and resp_expiry != machine_token_file.contract_expiry_datetime
    ):
        orig_token["machineTokenInfo"]["contractInfo"][
            "effectiveTo"
        ] = resp_expiry
        machine_token_file.write(orig_token)


def update_motd_messages(cfg: UAConfig) -> bool:
    """Emit human-readable status message used by motd.

    Used by /etc/update.motd.d/91-contract-ua-esm-status

    :param cfg: UAConfig instance for this environment.
    """
    is_attached_info = _is_attached(cfg)
    if not is_attached_info.is_attached:
        return False

    LOG.info("Updating Ubuntu Pro messages for MOTD.")
    motd_contract_status_msg_path = os.path.join(
        cfg.data_dir, "messages", MOTD_CONTRACT_STATUS_FILE_NAME
    )

    expiry_status = is_attached_info.contract_status
    remaining_days = is_attached_info.contract_remaining_days
    machine_token_file = machine_token.get_machine_token_file(cfg)

    if expiry_status in (
        ContractExpiryStatus.ACTIVE_EXPIRED_SOON.value,
        ContractExpiryStatus.EXPIRED_GRACE_PERIOD.value,
        ContractExpiryStatus.EXPIRED.value,
    ):
        update_contract_expiry(cfg)
        is_attached_info = _is_attached(cfg)
        expiry_status = is_attached_info.contract_status
        remaining_days = is_attached_info.contract_remaining_days

    if expiry_status in (
        ContractExpiryStatus.ACTIVE.value,
        ContractExpiryStatus.NONE.value,
    ):
        notices.remove(notices.Notice.CONTRACT_EXPIRED)
        system.ensure_file_absent(motd_contract_status_msg_path)
    elif expiry_status == ContractExpiryStatus.ACTIVE_EXPIRED_SOON.value:
        notices.remove(notices.Notice.CONTRACT_EXPIRED)
        system.write_file(
            motd_contract_status_msg_path,
            messages.CONTRACT_EXPIRES_SOON.pluralize(remaining_days).format(
                remaining_days=remaining_days,
            )
            + "\n\n",
        )
    elif expiry_status == ContractExpiryStatus.EXPIRED_GRACE_PERIOD.value:
        notices.remove(notices.Notice.CONTRACT_EXPIRED)
        grace_period_remaining = (
            defaults.CONTRACT_EXPIRY_GRACE_PERIOD_DAYS + remaining_days
        )
        exp_dt = machine_token_file.contract_expiry_datetime
        if exp_dt is None:
            exp_dt_str = "Unknown"
        else:
            exp_dt_str = exp_dt.strftime("%d %b %Y")
        system.write_file(
            motd_contract_status_msg_path,
            messages.CONTRACT_EXPIRED_GRACE_PERIOD.pluralize(
                remaining_days
            ).format(
                expired_date=exp_dt_str,
                remaining_days=grace_period_remaining,
            )
            + "\n\n",
        )
    elif expiry_status == ContractExpiryStatus.EXPIRED.value:
        notices.add(notices.Notice.CONTRACT_EXPIRED)

        service = "n/a"
        pkg_num = 0

        enabled_services_names = {
            s.name for s in _enabled_services(cfg).enabled_services
        }
        if system.is_current_series_active_esm():
            if "esm-infra" in enabled_services_names:
                service = "esm-infra"
                pkg_num = api_u_pro_packages_updates_v1(
                    cfg
                ).summary.num_esm_infra_updates
        elif system.is_current_series_lts():
            if "esm-apps" in enabled_services_names:
                service = "esm-apps"
                pkg_num = api_u_pro_packages_updates_v1(
                    cfg
                ).summary.num_esm_apps_updates

        if pkg_num == 0:
            system.write_file(
                motd_contract_status_msg_path,
                messages.CONTRACT_EXPIRED + "\n\n",
            )
        else:
            system.write_file(
                motd_contract_status_msg_path,
                messages.CONTRACT_EXPIRED_WITH_PKGS.pluralize(pkg_num).format(
                    pkg_num=pkg_num,
                    service=service,
                )
                + "\n\n",
            )

    return True


def refresh_motd():
    # If update-notifier is present, we might as well update
    # the package updates count related to MOTD
    if exists(UPDATE_NOTIFIER_MOTD_SCRIPT):
        # If this command fails, we shouldn't break the entire command,
        # since this command should already be triggered by
        # update-notifier apt hooks
        try:
            system.subp([UPDATE_NOTIFIER_MOTD_SCRIPT, "--force"])
        except Exception as exc:
            LOG.exception(exc)



import mock
import pytest

from uaclient.cli.config import show_subcommand

M_PATH = "uaclient.cli."


class TestActionConfigShow:
    @pytest.mark.parametrize(
        "optional_key",
        (
            None,
            "https_proxy",
            "http_proxy",
            "global_apt_http_proxy",
            "global_apt_https_proxy",
        ),
    )
    def test_show_values_and_limit_when_optional_key_provided(
        self, optional_key, FakeConfig, capsys
    ):
        cfg = FakeConfig()
        cfg.user_config.http_proxy = "http://http_proxy"
        cfg.user_config.https_proxy = "http://https_proxy"
        cfg.user_config.global_apt_http_proxy = "http://global_apt_http_proxy"
        cfg.user_config.global_apt_https_proxy = (
            "http://global_apt_https_proxy"
        )
        cfg.user_config.vulnerability_data_url_prefix = (
            "https://security-metadata.canonical.com/oval/"
        )
        args = mock.MagicMock(key=optional_key)
        show_subcommand.action(args, cfg=cfg)
        out, err = capsys.readouterr()
        if optional_key:
            assert "{key} http://{key}\n".format(key=optional_key) == out
        else:
            assert (
                """\
http_proxy                     http://http_proxy
https_proxy                    http://https_proxy
apt_http_proxy                 None
apt_https_proxy                None
ua_apt_http_proxy              None
ua_apt_https_proxy             None
global_apt_http_proxy          http://global_apt_http_proxy
global_apt_https_proxy         http://global_apt_https_proxy
update_messaging_timer         21600
metering_timer                 14400
apt_news                       True
apt_news_url                   https://motd.ubuntu.com/aptnews.json
vulnerability_data_url_prefix  https://security-metadata.canonical.com/oval/
lxd_guest_attach               off
"""
                == out
            )
        assert "" == err



import mock
import pytest

from uaclient.api import exceptions
from uaclient.api.u.pro.attach.guest.get_guest_token.v1 import (
    GetGuestTokenResult,
    _get_guest_token,
)
from uaclient.testing import helpers

M_PATH = "uaclient.api.u.pro.attach.guest.get_guest_token.v1."


class TestGetGuestTokenV1:
    @pytest.mark.parametrize(
        [
            "root",
            "attached",
            "machine_token_dict",
            "contract_id",
            "machine_id",
            "expected_exception",
            "expected_contract_calls",
        ],
        [
            (
                False,
                None,
                None,
                None,
                None,
                pytest.raises(exceptions.NonRootUserError),
                [],
            ),
            (
                True,
                False,
                None,
                None,
                None,
                pytest.raises(exceptions.UnattachedError),
                [],
            ),
            (
                True,
                True,
                None,
                None,
                None,
                pytest.raises(exceptions.UnattachedError),
                [],
            ),
            (
                True,
                True,
                {"machineToken": "fake_machine_token"},
                "contractId",
                "machineId",
                helpers.does_not_raise(),
                [
                    mock.call(
                        machine_token="fake_machine_token",
                        contract_id="contractId",
                        machine_id="machineId",
                    )
                ],
            ),
        ],
    )
    @mock.patch(M_PATH + "system.get_machine_id")
    @mock.patch(M_PATH + "machine_token.get_machine_token_file")
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "util.we_are_currently_root")
    @mock.patch(M_PATH + "contract.UAContractClient.get_guest_token")
    def test_get_guest_token(
        self,
        m_get_guest_token,
        m_we_are_currently_root,
        m_is_attached,
        m_get_machine_token_file,
        m_get_machine_id,
        root,
        attached,
        machine_token_dict,
        contract_id,
        machine_id,
        expected_exception,
        expected_contract_calls,
        FakeConfig,
    ):
        m_we_are_currently_root.return_value = root
        m_is_attached.return_value.is_attached = attached
        m_get_machine_token_file.return_value.machine_token = (
            machine_token_dict
        )
        m_get_machine_token_file.return_value.contract_id = contract_id
        m_get_machine_id.return_value = machine_id
        fake_guest_token = {
            "guestToken": "fake_guest_token",
            "id": "fake_id",
            "expires": "fake_expires",
        }
        m_get_guest_token.return_value = fake_guest_token
        expected_result = GetGuestTokenResult(
            guest_token=fake_guest_token["guestToken"],
            id=fake_guest_token["id"],
            expires=fake_guest_token["expires"],
        )
        with expected_exception:
            assert expected_result == _get_guest_token(FakeConfig())

        assert m_get_guest_token.call_args_list == expected_contract_calls



import mock
import pytest

from uaclient.cli.collect_logs import collect_logs_command
from uaclient.defaults import APPARMOR_PROFILES

M_PATH = "uaclient.cli."


class TestActionCollectLogs:
    @pytest.mark.parametrize(
        "series,extension", (("jammy", "list"), ("noble", "sources"))
    )
    @pytest.mark.parametrize("is_root", ((True), (False)))
    @mock.patch("uaclient.util.we_are_currently_root")
    @mock.patch(
        "glob.glob",
        return_value=[
            "/var/log/ubuntu-advantage.log",
            "/var/log/ubuntu-advantage.log.1",
        ],
    )
    @mock.patch("tarfile.open")
    @mock.patch("builtins.open")
    @mock.patch(M_PATH + "util.redact_sensitive_logs", return_value="test")
    # let's pretend all files exist
    @mock.patch("pathlib.Path.stat")
    @mock.patch("os.chown")
    @mock.patch("os.path.isfile", return_value=True)
    @mock.patch("shutil.copy")
    @mock.patch("uaclient.actions.status")
    @mock.patch("uaclient.system.write_file")
    @mock.patch("uaclient.system.load_file")
    @mock.patch("uaclient.system.subp", return_value=(None, None))
    @mock.patch("uaclient.log.get_user_log_file")
    @mock.patch("uaclient.log.get_all_user_log_files")
    @mock.patch("uaclient.system.get_release_info")
    def test_collect_logs(
        self,
        m_get_release_info,
        m_get_users,
        m_get_user,
        m_subp,
        _load_file,
        _write_file,
        m_status,
        m_shutilcopy,
        m_isfile,
        _chown,
        _stat,
        redact,
        _fopen,
        _tarfile,
        _glob,
        util_we_are_currently_root,
        is_root,
        series,
        extension,
        FakeConfig,
        tmpdir,
    ):
        m_status.return_value = {"user-id": ""}, 0
        m_get_release_info.return_value.series = series
        util_we_are_currently_root.return_value = is_root
        m_get_user.return_value = tmpdir.join("user-log").strpath
        m_get_users.return_value = [
            tmpdir.join("user1-log").strpath,
            tmpdir.join("user2-log").strpath,
        ]
        is_file_calls = 19 + len(APPARMOR_PROFILES)
        user_log_files = [mock.call(m_get_user())]
        if util_we_are_currently_root():
            user_log_files = [
                mock.call(m_get_users()[0]),
                mock.call(m_get_users()[1]),
            ]

        cfg = FakeConfig()
        collect_logs_command.action(mock.MagicMock(), cfg=cfg)

        assert m_subp.call_args_list == [
            mock.call(["cloud-id"], rcs=None),
            mock.call(["/snap/bin/canonical-livepatch", "status"], rcs=None),
            mock.call(["systemctl", "list-timers", "--all"], rcs=None),
            mock.call(
                [
                    "journalctl",
                    "--boot=0",
                    "-o",
                    "short-precise",
                    "-u",
                    "cloud-init-local.service",
                    "-u",
                    "cloud-init-config.service",
                    "-u",
                    "cloud-config.service",
                ],
                rcs=None,
            ),
            mock.call(
                [
                    "journalctl",
                    "-o",
                    "short-precise",
                    "-u",
                    "apt-news.service",
                    "-u",
                    "esm-cache.service",
                    "-u",
                    "ua-timer.service",
                    "-u",
                    "ua-auto-attach.service",
                    "-u",
                    "ua-reboot-cmds.service",
                    "-u",
                    "ubuntu-advantage.service",
                ],
                rcs=None,
            ),
            mock.call(["systemctl", "status", "apt-news.service"], rcs=[0, 3]),
            mock.call(
                ["systemctl", "status", "esm-cache.service"], rcs=[0, 3]
            ),
            mock.call(["systemctl", "status", "ua-timer.service"], rcs=[0, 3]),
            mock.call(["systemctl", "status", "ua-timer.timer"], rcs=[0, 3]),
            mock.call(
                ["systemctl", "status", "ua-auto-attach.path"], rcs=[0, 3]
            ),
            mock.call(
                ["systemctl", "status", "ua-auto-attach.service"], rcs=[0, 3]
            ),
            mock.call(
                ["systemctl", "status", "ua-reboot-cmds.service"], rcs=[0, 3]
            ),
            mock.call(
                ["systemctl", "status", "ubuntu-advantage.service"], rcs=[0, 3]
            ),
            mock.call(["journalctl", "-b", "-k", "--since=1 day ago"]),
        ]

        assert m_isfile.call_count == is_file_calls
        assert m_isfile.call_args_list == [
            mock.call("/etc/ubuntu-advantage/uaclient.conf"),
            mock.call(cfg.log_file),
            mock.call("/var/lib/ubuntu-advantage/jobs-status.json"),
            mock.call("/etc/cloud/build.info"),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-anbox-cloud.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-cc-eal.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-cis.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-esm-apps.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-esm-apps-legacy.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-esm-infra.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-esm-infra-legacy.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-fips.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-fips-updates.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-fips-preview.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-realtime-kernel.{}".format(
                    extension
                )
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-ros.{}".format(extension)
            ),
            mock.call(
                "/etc/apt/sources.list.d/ubuntu-ros-updates.{}".format(
                    extension
                )
            ),
            mock.call("/var/log/ubuntu-advantage.log"),
            mock.call("/var/log/ubuntu-advantage.log.1"),
            *[mock.call(f) for f in APPARMOR_PROFILES],
        ]
        # APPARMOR_PROFILES are not redacted
        assert redact.call_count == is_file_calls + len(user_log_files) - len(
            APPARMOR_PROFILES
        )
        assert m_shutilcopy.call_count == len(APPARMOR_PROFILES)



import textwrap
from typing import (  # noqa: F401
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from uaclient import apt, exceptions, messages, system, util
from uaclient.actions import attach_with_token, enable_entitlement_by_name
from uaclient.api.u.pro.attach.magic.initiate.v1 import _initiate
from uaclient.api.u.pro.attach.magic.revoke.v1 import (
    MagicAttachRevokeOptions,
    _revoke,
)
from uaclient.api.u.pro.attach.magic.wait.v1 import (
    MagicAttachWaitOptions,
    _wait,
)
from uaclient.api.u.pro.security.fix._common import (
    FixStatus,
    UnfixedPackage,
    status_message,
)
from uaclient.api.u.pro.security.fix._common.plan.v1 import (  # noqa: F401
    ESM_APPS_POCKET,
    ESM_INFRA_POCKET,
    STANDARD_UPDATES_POCKET,
    FixPlanAptUpgradeStep,
    FixPlanAttachStep,
    FixPlanEnableStep,
    FixPlanNoOpAlreadyFixedStep,
    FixPlanNoOpLivepatchFixStep,
    FixPlanNoOpStatus,
    FixPlanNoOpStep,
    FixPlanResult,
    FixPlanStep,
    FixPlanUSNResult,
    FixPlanWarning,
    FixPlanWarningFailUpdatingESMCache,
    FixPlanWarningPackageCannotBeInstalled,
    FixPlanWarningSecurityIssueNotFixed,
    NoOpAlreadyFixedData,
    NoOpLivepatchFixData,
    USNAdditionalData,
)
from uaclient.api.u.pro.security.fix.cve.plan.v1 import CVEFixPlanOptions
from uaclient.api.u.pro.security.fix.cve.plan.v1 import _plan as cve_plan
from uaclient.api.u.pro.security.fix.usn.plan.v1 import USNFixPlanOptions
from uaclient.api.u.pro.security.fix.usn.plan.v1 import _plan as usn_plan
from uaclient.api.u.pro.status.is_attached.v1 import (
    ContractExpiryStatus,
    _is_attached,
)
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.detach import action_detach
from uaclient.cli.parser import HelpCategory
from uaclient.clouds.identity import (
    CLOUD_TYPE_TO_TITLE,
    PRO_CLOUD_URLS,
    get_cloud_type,
)
from uaclient.config import UAConfig
from uaclient.defaults import PRINT_WRAP_WIDTH
from uaclient.entitlements import entitlement_factory
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    CanEnableFailure,
    UserFacingStatus,
)
from uaclient.files import notices
from uaclient.files.notices import Notice
from uaclient.messages.urls import PRO_HOME_PAGE
from uaclient.status import colorize_commands


class FixContext:
    def __init__(
        self,
        title: str,
        dry_run: bool,
        affected_pkgs: List[str],
        cfg: UAConfig,
    ):
        self.pkg_index = 0
        self.unfixed_pkgs = []  # type: List[UnfixedPackage]
        self.installed_pkgs = set()  # type: Set[str]
        self.fix_status = FixStatus.SYSTEM_NON_VULNERABLE
        self.title = title
        self.affected_pkgs = affected_pkgs
        self.dry_run = dry_run
        self.cfg = cfg
        self.should_print_pkg_header = True
        self.warn_package_cannot_be_installed = False
        self.fixed_by_livepatch = False

    def print_fix_header(self):
        if self.affected_pkgs:
            msg = messages.SECURITY_AFFECTED_PKGS.pluralize(
                len(self.affected_pkgs)
            ).format(
                count=len(self.affected_pkgs),
                pkgs=", ".join(sorted(self.affected_pkgs)),
            )
            print(
                textwrap.fill(
                    msg,
                    width=PRINT_WRAP_WIDTH,
                    subsequent_indent="    ",
                    replace_whitespace=False,
                )
            )

    def print_pkg_header(
        self,
        source_pkgs: List[str],
        status: str,
        pocket: Optional[str] = None,
    ):
        if self.should_print_pkg_header:
            print(
                _format_packages_message(
                    pkg_list=source_pkgs,
                    status=status,
                    pkg_index=self.pkg_index,
                    num_pkgs=len(self.affected_pkgs),
                    pocket_source=(
                        get_pocket_description(pocket) if pocket else None
                    ),
                )
            )

    def add_unfixed_packages(self, pkgs: List[str], unfixed_reason: str):
        for pkg in pkgs:
            self.unfixed_pkgs.append(
                UnfixedPackage(pkg=pkg, unfixed_reason=unfixed_reason)
            )


def print_cve_header(cve: FixPlanResult):
    lines = [
        "{issue}: {description}".format(
            issue=cve.title.upper(), description=cve.description
        ),
        " - https://ubuntu.com/security/{}".format(cve.title.upper()),
    ]

    print("\n".join(lines))


def print_usn_header(fix_plan: FixPlanUSNResult):
    target_usn = fix_plan.target_usn_plan
    lines = [
        "{issue}: {description}".format(
            issue=target_usn.title.upper(), description=target_usn.description
        ),
    ]

    additional_data = target_usn.additional_data
    if isinstance(additional_data, USNAdditionalData):
        if additional_data.associated_cves:
            lines.append(messages.SECURITY_FOUND_CVES)
            for cve in additional_data.associated_cves:
                lines.append(
                    " - {}".format(
                        messages.urls.SECURITY_CVE_PAGE.format(cve=cve)
                    )
                )
        elif additional_data.associated_launchpad_bugs:
            lines.append(messages.SECURITY_FOUND_LAUNCHPAD_BUGS)
            for lp_bug in additional_data.associated_launchpad_bugs:
                lines.append(" - " + lp_bug)

    print("\n".join(lines))


def fix_cve(security_issue: str, dry_run: bool, cfg: UAConfig):
    fix_plan = cve_plan(
        options=CVEFixPlanOptions(cves=[security_issue]), cfg=cfg
    )

    error = fix_plan.cves_data.cves[0].error
    if error and error.msg:
        raise exceptions.AnonymousUbuntuProError(
            named_msg=messages.NamedMessage(
                error.code or "unexpected-error", error.msg
            )
        )
    print_cve_header(fix_plan.cves_data.cves[0])
    print()

    status, _ = execute_fix_plan(fix_plan.cves_data.cves[0], dry_run, cfg)
    return status


def fix_usn(
    security_issue: str, dry_run: bool, no_related: bool, cfg: UAConfig
):
    fix_plan = usn_plan(
        options=USNFixPlanOptions(usns=[security_issue]), cfg=cfg
    )
    error = fix_plan.usns_data.usns[0].target_usn_plan.error
    if error and error.msg:
        raise exceptions.AnonymousUbuntuProError(
            named_msg=messages.NamedMessage(
                error.code or "unexpected-error", error.msg
            )
        )
    print_usn_header(fix_plan.usns_data.usns[0])

    print(
        "\n"
        + messages.SECURITY_FIXING_REQUESTED_USN.format(
            issue_id=security_issue
        )
    )

    target_usn_status, _ = execute_fix_plan(
        fix_plan.usns_data.usns[0].target_usn_plan,
        dry_run,
        cfg,
    )

    if target_usn_status not in (
        FixStatus.SYSTEM_NON_VULNERABLE,
        FixStatus.SYSTEM_NOT_AFFECTED,
    ):
        return target_usn_status

    related_usns_plan = fix_plan.usns_data.usns[0].related_usns_plan
    if not related_usns_plan or no_related:
        return target_usn_status

    print(
        "\n"
        + messages.SECURITY_RELATED_USNS.format(
            related_usns="\n- ".join(usn.title for usn in related_usns_plan)
        )
    )

    print("\n" + messages.SECURITY_FIXING_RELATED_USNS)
    related_usn_status = (
        {}
    )  # type: Dict[str, Tuple[FixStatus, List[UnfixedPackage]]]
    for related_usn_plan in related_usns_plan:
        print("- {}".format(related_usn_plan.title))
        related_usn_status[related_usn_plan.title] = execute_fix_plan(
            related_usn_plan,
            dry_run,
            cfg,
        )
        print()

    print(messages.SECURITY_USN_SUMMARY)
    _handle_fix_status_message(
        target_usn_status,
        security_issue,
        context=messages.FIX_ISSUE_CONTEXT_REQUESTED,
    )

    failure_on_related_usn = False
    for related_usn_plan in related_usns_plan:
        status, unfixed_pkgs = related_usn_status[related_usn_plan.title]
        _handle_fix_status_message(
            status,
            related_usn_plan.title,
            context=messages.FIX_ISSUE_CONTEXT_RELATED,
        )

        if status == FixStatus.SYSTEM_VULNERABLE_UNTIL_REBOOT:
            print(
                "- "
                + messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
                    operation="fix operation"
                )
            )
            failure_on_related_usn = True
        if status == FixStatus.SYSTEM_STILL_VULNERABLE:
            for unfixed_pkg in unfixed_pkgs:
                if unfixed_pkg.unfixed_reason:
                    print(
                        "  - {}: {}".format(
                            unfixed_pkg.pkg, unfixed_pkg.unfixed_reason
                        )
                    )
            failure_on_related_usn = True

    if failure_on_related_usn:
        print(
            "\n"
            + messages.SECURITY_RELATED_USN_ERROR.format(
                issue_id=security_issue
            )
        )

    return target_usn_status


def _format_packages_message(
    pkg_list: List[str],
    status: str,
    pkg_index: int,
    num_pkgs: int,
    pocket_source: Optional[str] = None,
) -> str:
    """Format the packages and status to an user friendly message."""
    if not pkg_list:
        return ""

    msg_index = []
    src_pkgs = []
    for src_pkg in pkg_list:
        pkg_index += 1
        msg_index.append("{}/{}".format(pkg_index, num_pkgs))
        src_pkgs.append(src_pkg)

    msg_header = textwrap.fill(
        "{} {}:".format(
            "(" + ", ".join(msg_index) + ")", ", ".join(sorted(src_pkgs))
        ),
        width=PRINT_WRAP_WIDTH,
        subsequent_indent="    ",
    )
    return "{}\n{}".format(msg_header, status_message(status, pocket_source))


def _run_ua_attach(cfg: UAConfig, token: str) -> bool:
    """Attach to an Ubuntu Pro subscription with a given token.

    :return: True if attach performed without errors.
    """
    print(colorize_commands([["pro", "attach", token]]))
    try:
        attach_with_token(cfg, token=token, allow_enable=True)
        return True
    except exceptions.UbuntuProError as err:
        print(err.msg)
        return False


def _inform_ubuntu_pro_existence_if_applicable() -> None:
    """Alert the user when running Pro on cloud with PRO support."""
    cloud_type, _ = get_cloud_type()
    if cloud_type in PRO_CLOUD_URLS.keys():
        print(
            messages.SECURITY_USE_PRO_TMPL.format(
                title=CLOUD_TYPE_TO_TITLE.get(cloud_type),
                cloud_specific_url=PRO_CLOUD_URLS.get(cloud_type),
            )
        )


def _perform_magic_attach(cfg: UAConfig):
    print(messages.CLI_MAGIC_ATTACH_INIT)
    initiate_resp = _initiate(cfg=cfg)
    print(
        "\n"
        + messages.CLI_MAGIC_ATTACH_SIGN_IN.format(
            user_code=initiate_resp.user_code
        )
    )

    wait_options = MagicAttachWaitOptions(magic_token=initiate_resp.token)

    try:
        wait_resp = _wait(options=wait_options, cfg=cfg)
    except exceptions.MagicAttachTokenError as e:
        print(messages.CLI_MAGIC_ATTACH_FAILED)

        revoke_options = MagicAttachRevokeOptions(
            magic_token=initiate_resp.token
        )
        _revoke(options=revoke_options, cfg=cfg)
        raise e

    print("\n" + messages.CLI_MAGIC_ATTACH_PROCESSING)
    return _run_ua_attach(cfg, wait_resp.contract_token)


def _prompt_for_attach(cfg: UAConfig) -> bool:
    """Prompt for attach to a subscription or token.

    :return: True if attach performed.
    """
    _inform_ubuntu_pro_existence_if_applicable()
    print(messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION)
    choice = util.prompt_choices(
        messages.SECURITY_FIX_ATTACH_PROMPT,
        valid_choices=["s", "a", "c"],
    )
    if choice == "c":
        return False
    if choice == "s":
        return _perform_magic_attach(cfg)
    if choice == "a":
        print(messages.PROMPT_ENTER_TOKEN)
        token = input("> ")
        return _run_ua_attach(cfg, token)

    return True


def _format_unfixed_packages_msg(unfixed_pkgs: List[str]) -> str:
    """Format the list of unfixed packages into an message.

    :returns: A string containing the message output for the unfixed
              packages.
    """
    num_pkgs_unfixed = len(unfixed_pkgs)
    return textwrap.fill(
        messages.SECURITY_PKG_STILL_AFFECTED.pluralize(
            num_pkgs_unfixed
        ).format(
            num_pkgs=num_pkgs_unfixed,
            pkgs=", ".join(sorted(unfixed_pkgs)),
        ),
        width=PRINT_WRAP_WIDTH,
        subsequent_indent="    ",
    )


def _check_subscription_is_expired(cfg: UAConfig, dry_run: bool) -> bool:
    """Check if the Ubuntu Pro subscription is expired.

    :returns: True if subscription is expired and not renewed.
    """
    contract_expiry_status = _is_attached(cfg).contract_status
    if (
        contract_expiry_status
        and contract_expiry_status == ContractExpiryStatus.EXPIRED.value
    ):
        if dry_run:
            print(messages.SECURITY_DRY_RUN_UA_EXPIRED_SUBSCRIPTION)
            return False
        return True

    return False


def _prompt_for_new_token(cfg: UAConfig) -> bool:
    """Prompt for attach a new subscription token to the user.

    :return: True if attach performed.
    """
    import argparse

    _inform_ubuntu_pro_existence_if_applicable()
    print(messages.SECURITY_UPDATE_NOT_INSTALLED_EXPIRED)
    choice = util.prompt_choices(
        messages.SECURITY_FIX_RENEW_PROMPT.format(url=PRO_HOME_PAGE),
        valid_choices=["r", "c"],
    )
    if choice == "r":
        print(messages.PROMPT_EXPIRED_ENTER_TOKEN)
        token = input("> ")
        print(colorize_commands([["pro", "detach"]]))
        action_detach(argparse.Namespace(assume_yes=True, format="cli"), cfg)
        return _run_ua_attach(cfg, token)

    return False


def _prompt_for_enable(cfg: UAConfig, service: str) -> bool:
    """Prompt for enable a pro service.

    :return: True if enable performed.
    """
    print(messages.SECURITY_SERVICE_DISABLED.format(service=service))
    choice = util.prompt_choices(
        messages.SECURITY_FIX_ENABLE_PROMPT.format(service=service),
        valid_choices=["e", "c"],
    )

    if choice == "e":
        print(colorize_commands([["pro", "enable", service]]))
        ret, reason = enable_entitlement_by_name(cfg=cfg, name=service)

        if (
            not ret
            and reason is not None
            and isinstance(reason, CanEnableFailure)
        ):
            if reason.message is not None:
                print(reason.message.msg)

        return ret

    return False


def _handle_subscription_for_required_service(
    service: str, cfg: UAConfig, dry_run: bool
) -> bool:
    """
    Verify if the Ubuntu Pro subscription has the required service enabled.
    """
    ent = entitlement_factory(cfg=cfg, name=service)
    if ent:
        ent_status, _ = ent.user_facing_status()

        if ent_status == UserFacingStatus.ACTIVE:
            return True

        applicability_status, _ = ent.applicability_status()
        if applicability_status == ApplicabilityStatus.APPLICABLE:
            if dry_run:
                print(
                    "\n"
                    + messages.SECURITY_DRY_RUN_UA_SERVICE_NOT_ENABLED.format(
                        service=ent.name
                    )
                )
                return True

            if _prompt_for_enable(cfg, ent.name):
                return True
            else:
                print(
                    messages.SECURITY_UA_SERVICE_NOT_ENABLED.format(
                        service=ent.name
                    )
                )

        else:
            print(
                messages.SECURITY_UA_SERVICE_NOT_ENTITLED.format(
                    service=ent.name
                )
            )

    return False


def _handle_fix_status_message(
    status: FixStatus, issue_id: str, context: str = ""
):
    if status == FixStatus.SYSTEM_NON_VULNERABLE:
        if context:
            msg = messages.SECURITY_ISSUE_RESOLVED_ISSUE_CONTEXT.format(
                issue=issue_id, context=context
            )
        else:
            msg = messages.SECURITY_ISSUE_RESOLVED.format(issue=issue_id)
        print(util.handle_unicode_characters(msg))
    elif status == FixStatus.SYSTEM_NOT_AFFECTED:
        if context:
            msg = messages.SECURITY_ISSUE_UNAFFECTED_ISSUE_CONTEXT.format(
                issue=issue_id, context=context
            )
        else:
            msg = messages.SECURITY_ISSUE_UNAFFECTED.format(issue=issue_id)
        print(util.handle_unicode_characters(msg))
    elif status == FixStatus.SYSTEM_VULNERABLE_UNTIL_REBOOT:
        if context:
            msg = messages.SECURITY_ISSUE_NOT_RESOLVED_ISSUE_CONTEXT.format(
                issue=issue_id, context=context
            )
        else:
            msg = messages.SECURITY_ISSUE_NOT_RESOLVED.format(issue=issue_id)
        print(util.handle_unicode_characters(msg))
    else:
        if context:
            msg = messages.SECURITY_ISSUE_NOT_RESOLVED_ISSUE_CONTEXT.format(
                issue=issue_id, context=context
            )
        else:
            msg = messages.SECURITY_ISSUE_NOT_RESOLVED.format(issue=issue_id)
        print(util.handle_unicode_characters(msg))


def get_pocket_description(pocket: str):
    if pocket == STANDARD_UPDATES_POCKET:
        return messages.SECURITY_UBUNTU_STANDARD_UPDATES_POCKET
    elif pocket == ESM_INFRA_POCKET:
        return messages.SECURITY_UA_INFRA_POCKET
    elif pocket == ESM_APPS_POCKET:
        return messages.SECURITY_UA_APPS_POCKET
    else:
        return pocket


def _execute_package_cannot_be_installed_step(
    fix_context: FixContext,
    step: FixPlanWarningPackageCannotBeInstalled,
):
    fix_context.print_pkg_header(
        source_pkgs=step.data.related_source_packages,
        status="released",
        pocket=step.data.pocket,
    )
    fix_context.should_print_pkg_header = False

    warn_msg = messages.FIX_CANNOT_INSTALL_PACKAGE.format(
        package=step.data.binary_package,
        version=step.data.binary_package_version,
    )
    print("- " + warn_msg)

    fix_context.add_unfixed_packages(
        pkgs=[step.data.source_package], unfixed_reason=warn_msg
    )

    fix_context.warn_package_cannot_be_installed = True
    fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE


def _execute_security_issue_not_fixed_step(
    fix_context: FixContext, step: FixPlanWarningSecurityIssueNotFixed
):
    fix_context.print_pkg_header(
        source_pkgs=step.data.source_packages,
        status=step.data.status,
    )
    fix_context.pkg_index += len(step.data.source_packages)

    fix_context.add_unfixed_packages(
        pkgs=step.data.source_packages,
        unfixed_reason=status_message(step.data.status),
    )
    fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE


def _execute_fail_updating_esm_cache_step(
    fix_context: FixContext, step: FixPlanWarningFailUpdatingESMCache
):
    if util.we_are_currently_root():
        print(messages.CLI_FIX_FAIL_UPDATING_ESM_CACHE)
    else:
        print("\n" + messages.CLI_FIX_FAIL_UPDATING_ESM_CACHE_NON_ROOT + "\n")


def _execute_apt_upgrade_step(
    fix_context: FixContext,
    step: FixPlanAptUpgradeStep,
):
    fix_context.print_pkg_header(
        source_pkgs=step.data.source_packages,
        status="released",
        pocket=step.data.pocket,
    )
    fix_context.pkg_index += len(step.data.source_packages)

    if not step.data.binary_packages:
        if not fix_context.warn_package_cannot_be_installed:
            print(messages.SECURITY_UPDATE_INSTALLED)
        fix_context.fix_status = FixStatus.SYSTEM_NON_VULNERABLE
        return

    if not util.we_are_currently_root() and not fix_context.dry_run:
        print(messages.SECURITY_APT_NON_ROOT)
        fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE
        fix_context.add_unfixed_packages(
            pkgs=step.data.source_packages,
            unfixed_reason=messages.SECURITY_APT_NON_ROOT,
        )
        return

    print(
        colorize_commands(
            [
                ["apt", "update", "&&"]
                + ["apt", "install", "--only-upgrade", "-y"]
                + sorted(step.data.binary_packages)
            ]
        )
    )

    if fix_context.dry_run:
        fix_context.fix_status = FixStatus.SYSTEM_NON_VULNERABLE
        return

    try:
        apt.run_apt_update_command()
        apt.run_apt_command(
            cmd=["apt-get", "install", "--only-upgrade", "-y"]
            + step.data.binary_packages,
            override_env_vars={"DEBIAN_FRONTEND": "noninteractive"},
        )
    except Exception as e:
        msg = getattr(e, "msg", str(e))
        print(msg)
        fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE
        fix_context.add_unfixed_packages(
            pkgs=step.data.source_packages,
            unfixed_reason=msg,
        )
        return

    fix_context.fix_status = FixStatus.SYSTEM_NON_VULNERABLE
    fix_context.should_print_pkg_header = True
    fix_context.installed_pkgs.update(step.data.binary_packages)


def _execute_attach_step(
    fix_context: FixContext,
    step: FixPlanAttachStep,
):
    pocket = (
        ESM_INFRA_POCKET
        if step.data.required_service == "esm-infra"
        else ESM_APPS_POCKET
    )
    fix_context.print_pkg_header(
        source_pkgs=step.data.source_packages,
        status="released",
        pocket=pocket,
    )

    fix_context.should_print_pkg_header = False
    if not _is_attached(fix_context.cfg).is_attached:
        if fix_context.dry_run:
            print("\n" + messages.SECURITY_DRY_RUN_UA_NOT_ATTACHED)
        else:
            if not _prompt_for_attach(fix_context.cfg):
                fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE
                fix_context.add_unfixed_packages(
                    pkgs=step.data.source_packages,
                    unfixed_reason=messages.SECURITY_UA_SERVICE_REQUIRED.format(  # noqa
                        service=step.data.required_service
                    ),
                )
                return
    elif _check_subscription_is_expired(
        cfg=fix_context.cfg, dry_run=fix_context.dry_run
    ):
        if fix_context.dry_run:
            print(messages.SECURITY_DRY_RUN_UA_EXPIRED_SUBSCRIPTION)
        elif not _prompt_for_new_token(fix_context.cfg):
            fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE
            fix_context.add_unfixed_packages(
                pkgs=step.data.source_packages,
                unfixed_reason=messages.SECURITY_UA_SERVICE_WITH_EXPIRED_SUB.format(  # noqa
                    service=step.data.required_service
                ),
            )
            return

    fix_context.fix_status = FixStatus.SYSTEM_NON_VULNERABLE


def _execute_enable_step(
    fix_context: FixContext,
    step: FixPlanEnableStep,
):
    pocket = (
        ESM_INFRA_POCKET
        if step.data.service == "esm-infra"
        else ESM_APPS_POCKET
    )
    fix_context.print_pkg_header(
        source_pkgs=step.data.source_packages,
        status="released",
        pocket=pocket,
    )
    fix_context.should_print_pkg_header = False

    if not _handle_subscription_for_required_service(  # noqa
        step.data.service,
        fix_context.cfg,
        fix_context.dry_run,
    ):
        fix_context.add_unfixed_packages(
            pkgs=step.data.source_packages,
            unfixed_reason=messages.SECURITY_UA_SERVICE_NOT_ENABLED_SHORT.format(  # noqa
                service=step.data.service
            ),
        )
        fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE
        return

    return FixStatus.SYSTEM_NON_VULNERABLE


def _execute_noop_not_affected_step(
    fix_context: FixContext, step: FixPlanNoOpStep
):
    if step.data.status == FixPlanNoOpStatus.NOT_AFFECTED.value:
        print(messages.SECURITY_NO_AFFECTED_PKGS)
        fix_context.fix_status = FixStatus.SYSTEM_NOT_AFFECTED


def _execute_noop_fixed_by_livepatch_step(
    fix_context: FixContext, step: FixPlanNoOpLivepatchFixStep
):
    if isinstance(step.data, NoOpLivepatchFixData):
        print(
            messages.CVE_FIXED_BY_LIVEPATCH.format(
                issue=fix_context.title,
                version=step.data.patch_version,
            )
        )
        fix_context.fixed_by_livepatch = True


def _execute_noop_already_fixed_step(
    fix_context: FixContext, step: FixPlanNoOpAlreadyFixedStep
):
    if isinstance(step.data, NoOpAlreadyFixedData):
        fix_context.print_pkg_header(
            source_pkgs=step.data.source_packages,
            status="released",
            pocket=step.data.pocket,
        )
        print(messages.SECURITY_UPDATE_INSTALLED)
        fix_context.pkg_index += len(step.data.source_packages)


def execute_fix_plan(
    fix_plan: FixPlanResult, dry_run: bool, cfg: UAConfig
) -> Tuple[FixStatus, List[UnfixedPackage]]:
    full_plan = [
        *fix_plan.plan,
        *fix_plan.warnings,
    ]  # type: List[Union[FixPlanStep, FixPlanWarning]]

    fix_context = FixContext(
        title=fix_plan.title,
        dry_run=dry_run,
        affected_pkgs=fix_plan.affected_packages or [],
        cfg=cfg,
    )
    fix_context.print_fix_header()

    for step in sorted(full_plan, key=lambda x: x.order):
        if isinstance(step, FixPlanWarningPackageCannotBeInstalled):
            _execute_package_cannot_be_installed_step(fix_context, step)
        if isinstance(step, FixPlanWarningSecurityIssueNotFixed):
            _execute_security_issue_not_fixed_step(fix_context, step)
        if isinstance(step, FixPlanWarningFailUpdatingESMCache):
            _execute_fail_updating_esm_cache_step(fix_context, step)
        if isinstance(step, FixPlanAptUpgradeStep):
            _execute_apt_upgrade_step(fix_context, step)

            if fix_context.fix_status != FixStatus.SYSTEM_NON_VULNERABLE:
                break
        if isinstance(step, FixPlanAttachStep):
            _execute_attach_step(fix_context, step)

            if fix_context.fix_status != FixStatus.SYSTEM_NON_VULNERABLE:
                break
        if isinstance(step, FixPlanEnableStep):
            _execute_enable_step(fix_context, step)

            if fix_context.fix_status != FixStatus.SYSTEM_NON_VULNERABLE:
                break

        if isinstance(step, FixPlanNoOpStep):
            _execute_noop_not_affected_step(fix_context, step)
        if isinstance(step, FixPlanNoOpLivepatchFixStep):
            _execute_noop_fixed_by_livepatch_step(fix_context, step)
        if isinstance(step, FixPlanNoOpAlreadyFixedStep):
            _execute_noop_already_fixed_step(fix_context, step)

    print()
    if fix_context.unfixed_pkgs:
        print(
            _format_unfixed_packages_msg(
                list(
                    set(
                        [
                            unfixed_pkg.pkg
                            for unfixed_pkg in fix_context.unfixed_pkgs
                        ]
                    )
                )
            )
        )
        fix_context.fix_status = FixStatus.SYSTEM_STILL_VULNERABLE

    if (
        fix_context.fix_status == FixStatus.SYSTEM_NON_VULNERABLE
        and system.should_reboot(installed_pkgs=fix_context.installed_pkgs)
    ):
        fix_context.fix_status = FixStatus.SYSTEM_VULNERABLE_UNTIL_REBOOT
        reboot_msg = messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
            operation="fix operation"
        )
        print(reboot_msg)
        notices.add(
            Notice.ENABLE_REBOOT_REQUIRED,
            operation="fix operation",
        )

    if not fix_context.fixed_by_livepatch:
        _handle_fix_status_message(fix_context.fix_status, fix_plan.title)

    return (fix_context.fix_status, fix_context.unfixed_pkgs)


@cli_util.assert_vulnerability_issue_valid(cmd="fix")
def action_fix(args, *, cfg, **kwargs):
    if args.dry_run:
        print(messages.SECURITY_DRY_RUN_WARNING)

    if "cve" in args.security_issue.lower():
        status = fix_cve(args.security_issue, args.dry_run, cfg)
    else:
        status = fix_usn(
            args.security_issue, args.dry_run, args.no_related, cfg
        )

    return status.exit_code


fix_command = ProCommand(
    "fix",
    help=messages.CLI_ROOT_FIX,
    description=messages.CLI_FIX_DESC,
    action=action_fix,
    help_category=HelpCategory.SECURITY,
    preserve_description=True,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument("security_issue", help=messages.CLI_FIX_ISSUE),
                ProArgument(
                    "--dry-run",
                    help=messages.CLI_FIX_DRY_RUN,
                    action="store_true",
                ),
                ProArgument(
                    "--no-related",
                    help=messages.CLI_FIX_NO_RELATED,
                    action="store_true",
                ),
            ]
        )
    ],
)



import mock
import pytest

from uaclient import exceptions, messages
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.entitlements.landscape import LandscapeEntitlement


class TestLandscapeEntitlement:
    @pytest.mark.parametrize(
        [
            "assume_yes",
            "extra_args",
            "subp_sideeffect",
            "expected_subp_calls",
            "expected_result",
        ],
        [
            (
                False,
                None,
                None,
                [mock.call(["landscape-config"], pipe_stdouterr=False)],
                True,
            ),
            (
                False,
                ["extra"],
                None,
                [
                    mock.call(
                        ["landscape-config", "extra"], pipe_stdouterr=False
                    )
                ],
                True,
            ),
            (
                True,
                None,
                None,
                [
                    mock.call(
                        ["landscape-config", "--silent"], pipe_stdouterr=True
                    )
                ],
                True,
            ),
            (
                True,
                ["extra"],
                None,
                [
                    mock.call(
                        ["landscape-config", "extra", "--silent"],
                        pipe_stdouterr=True,
                    )
                ],
                True,
            ),
            (
                True,
                ["--silent", "extra"],
                None,
                [
                    mock.call(
                        ["landscape-config", "--silent", "extra"],
                        pipe_stdouterr=True,
                    )
                ],
                True,
            ),
            (
                False,
                None,
                exceptions.ProcessExecutionError("test"),
                [mock.call(["landscape-config"], pipe_stdouterr=False)],
                False,
            ),
        ],
    )
    @mock.patch("uaclient.system.subp")
    def test_perform_enable(
        self,
        m_subp,
        assume_yes,
        extra_args,
        subp_sideeffect,
        expected_subp_calls,
        expected_result,
        FakeConfig,
    ):
        m_subp.side_effect = subp_sideeffect
        landscape = LandscapeEntitlement(FakeConfig(), extra_args=extra_args)
        mock_progress = mock.MagicMock()
        mock_progress.is_interactive.return_value = not assume_yes
        assert expected_result == landscape._perform_enable(mock_progress)
        assert expected_subp_calls == m_subp.call_args_list

    @pytest.mark.parametrize(
        [
            "subp_sideeffect",
            "expected_subp_calls",
            "expected_result",
        ],
        [
            (
                None,
                [mock.call(["landscape-config", "--disable"])],
                True,
            ),
            (
                exceptions.ProcessExecutionError("test"),
                [mock.call(["landscape-config", "--disable"])],
                True,
            ),
        ],
    )
    @mock.patch("uaclient.system.subp")
    def test_perform_disable(
        self,
        m_subp,
        subp_sideeffect,
        expected_subp_calls,
        expected_result,
        FakeConfig,
    ):
        m_subp.side_effect = subp_sideeffect
        landscape = LandscapeEntitlement(FakeConfig())
        assert expected_result == landscape._perform_disable(mock.MagicMock())
        assert expected_subp_calls == m_subp.call_args_list

    @pytest.mark.parametrize(
        [
            "are_required_packages_installed",
            "is_systemd_unit_active",
            "expected_result",
        ],
        [
            (
                False,
                False,
                (
                    ApplicationStatus.DISABLED,
                    messages.LANDSCAPE_SERVICE_NOT_ACTIVE,
                ),
            ),
            (
                False,
                True,
                (
                    ApplicationStatus.DISABLED,
                    messages.LANDSCAPE_SERVICE_NOT_ACTIVE,
                ),
            ),
            (
                True,
                False,
                (
                    ApplicationStatus.DISABLED,
                    messages.LANDSCAPE_SERVICE_NOT_ACTIVE,
                ),
            ),
            (
                True,
                True,
                (ApplicationStatus.ENABLED, None),
            ),
        ],
    )
    @mock.patch("uaclient.system.is_systemd_unit_active")
    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement.are_required_packages_installed"  # noqa: E501
    )
    def test_application_status(
        self,
        m_are_required_packages_installed,
        m_is_systemd_unit_active,
        are_required_packages_installed,
        is_systemd_unit_active,
        expected_result,
        FakeConfig,
    ):
        m_are_required_packages_installed.return_value = (
            are_required_packages_installed
        )
        m_is_systemd_unit_active.return_value = is_systemd_unit_active
        landscape = LandscapeEntitlement(FakeConfig())
        assert expected_result == landscape.application_status()

    @pytest.mark.parametrize(
        [
            "we_are_currently_root",
            "subp_sideeffect",
            "expected_subp_calls",
            "expected_result",
        ],
        [
            (
                True,
                exceptions.ProcessExecutionError("test"),
                [mock.call(mock.ANY)],
                (True, messages.LANDSCAPE_NOT_REGISTERED),
            ),
            (True, None, [mock.call(mock.ANY)], (False, None)),
            (False, None, [], (False, None)),
        ],
    )
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_enabled_warning_status(
        self,
        m_we_are_currently_root,
        m_subp,
        we_are_currently_root,
        subp_sideeffect,
        expected_subp_calls,
        expected_result,
        FakeConfig,
    ):
        m_we_are_currently_root.return_value = we_are_currently_root
        m_subp.side_effect = subp_sideeffect
        landscape = LandscapeEntitlement(FakeConfig())
        assert expected_result == landscape.enabled_warning_status()
        assert expected_subp_calls == m_subp.call_args_list



import abc
import datetime
import enum
import json
import os
from collections import defaultdict
from functools import lru_cache
from typing import Any, Dict, List, NamedTuple, Optional
from urllib.parse import urljoin

from uaclient import apt, exceptions, http, system, util
from uaclient.api.u.pro.security.fix._common import (
    query_installed_source_pkg_versions,
)
from uaclient.api.u.pro.status.enabled_services.v1 import _enabled_services
from uaclient.config import UAConfig
from uaclient.data_types import (
    DataObject,
    Field,
    FloatDataValue,
    StringDataValue,
)
from uaclient.defaults import (
    VULNERABILITY_CACHE_PATH,
    VULNERABILITY_DATA_CACHE,
    VULNERABILITY_DATA_TMPL,
    VULNERABILITY_DPKG_STATUS_DATE_CACHE,
    VULNERABILITY_ETAG_CACHE,
    VULNERABILITY_RESULT_CACHE,
)
from uaclient.entitlements.fips import FIPSEntitlement, FIPSUpdatesEntitlement
from uaclient.files.data_types import DataObjectFile
from uaclient.files.files import UAFile


class VulnerabilityCacheETag(DataObject):
    fields = [Field("etag", StringDataValue)]

    def __init__(self, etag):
        self.etag = etag


class VulnerabilityDpkgCacheDate(DataObject):
    fields = [Field("dpkg_status_date", FloatDataValue)]

    def __init__(self, dpkg_status_date: float):
        self.dpkg_status_date = dpkg_status_date


@enum.unique
class VulnerabilityStatus(enum.Enum):
    """
    An enum to represent the status of a vulnerability
    """

    NO_FIX_AVAILABLE = "no"
    PARTIAL_FIX_AVAILABLE = "partial"
    FULL_FIX_AVAILABLE = "yes"


class VulnerabilityData:

    def __init__(
        self,
        cfg: UAConfig,
        series: Optional[str] = None,
    ):
        self.cfg = cfg
        self.series = series or system.get_release_info().series
        self._etag = None  # type: Optional[str]
        self._refreshed = False

    @property
    def refreshed(self):
        return self._refreshed

    def _get_cache_data_path(self):
        return os.path.join(
            VULNERABILITY_CACHE_PATH, self.series, VULNERABILITY_DATA_CACHE
        )

    def _get_etag_cache_file(self):
        return DataObjectFile(
            data_object_cls=VulnerabilityCacheETag,
            ua_file=UAFile(
                name=VULNERABILITY_ETAG_CACHE,
                directory=os.path.join(VULNERABILITY_CACHE_PATH, self.series),
                private=False,
            ),
        )

    def _save_cache_data(self, json_data: Dict[str, Any]):
        system.write_file(self._get_cache_data_path(), json.dumps(json_data))

    def _save_etag_cache(self, cache_etag_file: DataObjectFile, etag: str):
        cache_etag_file.write(VulnerabilityCacheETag(etag=etag))

    def _get_etag(self):
        if not self._etag:
            etag_file = self._get_etag_cache_file()
            etag_data = etag_file.read()

            if etag_data:
                self._etag = etag_data.etag

        return self._etag

    def _get_cache_data(self):
        return json.loads(system.load_file(self._get_cache_data_path()))

    def _get_data_url(self):
        data_name = self.series

        enabled_services_names = [
            s.name for s in _enabled_services(self.cfg).enabled_services
        ]
        if FIPSEntitlement.name in enabled_services_names:
            data_name = "fips_{}".format(self.series)
        elif FIPSUpdatesEntitlement.name in enabled_services_names:
            data_name = "fips-updates_{}".format(self.series)

        data_file = VULNERABILITY_DATA_TMPL.format(series=data_name)
        return urljoin(self.cfg.vulnerability_data_url_prefix, data_file)

    def get_published_date(self):
        vulnerability_json_data = self.get()
        return vulnerability_json_data["published_at"]

    def get(self):
        last_etag = self._get_etag()

        try:
            data, etag = http.download_xz_file_from_url(
                cfg=self.cfg, url=self._get_data_url(), etag=last_etag
            )
            self._refreshed = True
        except exceptions.ETagUnchanged:
            return self._get_cache_data()

        json_data = json.loads(data.decode("utf-8"))

        if util.we_are_currently_root():
            self._save_cache_data(json_data)
            if etag:
                self._save_etag_cache(self._get_etag_cache_file(), etag)

        return json_data


def _get_vulnerability_fix_status(
    affected_packages: List[Dict[str, Optional[str]]],
) -> VulnerabilityStatus:
    vulnerability_status = VulnerabilityStatus.NO_FIX_AVAILABLE
    num_fixes = 0
    for pkg in affected_packages:
        if pkg.get("fix_version") is not None:
            num_fixes += 1

    if num_fixes == len(affected_packages):
        vulnerability_status = VulnerabilityStatus.FULL_FIX_AVAILABLE
    elif num_fixes != 0:
        vulnerability_status = VulnerabilityStatus.PARTIAL_FIX_AVAILABLE

    return vulnerability_status


VulnerabilityParserResult = NamedTuple(
    "VulnerabilityParserResult",
    [
        ("vulnerability_data_published_at", Optional[datetime.datetime]),
        ("vulnerabilities_info", Dict[str, Dict[str, Any]]),
    ],
)


class VulnerabilitiesAlreadyFixed:
    def __init__(self):
        self._vulns = defaultdict(set)
        self.priority_counter = defaultdict(
            lambda: defaultdict(int)
        )  # type: Dict[str, Dict[str, int]]

    def add_vulnerability(
        self,
        vuln_name: str,
        vuln_pocket: str,
        vuln_priority: Optional[str] = None,
    ):
        if vuln_name not in self._vulns[vuln_pocket]:
            self._vulns[vuln_pocket].add(vuln_name)

            if vuln_priority:
                self.priority_counter[vuln_pocket][vuln_priority] += 1

    def to_dict(self):
        dict_repr = {
            "count": {},
            "info": {},
        }  # type: Dict[str, Dict[str, Any]]
        for pocket, vulns in self._vulns.items():
            dict_repr["count"][pocket] = len(vulns)
            dict_repr["info"][pocket] = dict(self.priority_counter[pocket])

        return dict_repr


class VulnerabilityParser(metaclass=abc.ABCMeta):
    vulnerability_type = None  # type: str

    @abc.abstractmethod
    def get_package_vulnerabilities(
        self, affected_pkg: Dict[str, Any]
    ) -> Dict[str, Any]:
        pass

    @abc.abstractmethod
    def _post_process_vulnerability_info(
        self,
        vulnerability_info: Dict[str, Any],
        vulnerabilities_data: Dict[str, Any],
    ) -> Dict[str, Any]:
        pass

    def _add_new_vulnerability(
        self,
        packages: Dict[str, Any],
        bin_pkg_name: str,
        bin_pkg_version: str,
        vuln_name: str,
    ):
        packages[bin_pkg_name] = {
            "current_version": bin_pkg_version,
            self.vulnerability_type: [],
        }

    def _add_unfixable_vulnerability(
        self,
        packages: Dict[str, Any],
        bin_pkg_name: str,
        bin_pkg_version: str,
        vuln_name: str,
        vuln_pkg_status: str,
    ):
        if bin_pkg_name not in packages:
            self._add_new_vulnerability(
                packages=packages,
                bin_pkg_name=bin_pkg_name,
                bin_pkg_version=bin_pkg_version,
                vuln_name=vuln_name,
            )

        packages[bin_pkg_name][self.vulnerability_type].append(
            {
                "name": vuln_name,
                "fix_version": None,
                "fix_status": vuln_pkg_status,
                "fix_origin": None,
            }
        )

    def _add_fixable_vulnerability(
        self,
        packages: Dict[str, Any],
        bin_pkg_name: str,
        bin_pkg_version: str,
        vuln_name: str,
        vuln_pkg_status: str,
        vuln_bin_fix_version: str,
        vuln_pocket: str,
    ):
        if bin_pkg_name not in packages:
            self._add_new_vulnerability(
                packages=packages,
                bin_pkg_name=bin_pkg_name,
                bin_pkg_version=bin_pkg_version,
                vuln_name=vuln_name,
            )

        packages[bin_pkg_name][self.vulnerability_type].append(
            {
                "name": vuln_name,
                "fix_version": vuln_bin_fix_version,
                "fix_status": vuln_pkg_status,
                "fix_origin": vuln_pocket,
            }
        )

    def _add_vulnerability_info(
        self,
        vuln_name: str,
        vulnerabilities: Dict[str, Any],
        vuln_info: Dict[str, Any],
        vulns_data: Dict[str, Any],
    ):
        if vuln_name not in vulnerabilities:
            vulnerabilities[vuln_name] = self._post_process_vulnerability_info(
                vulnerability_info=vuln_info,
                vulnerabilities_data=vulns_data,
            )

    def is_vulnerability_not_fixable(
        self,
        vuln_source_fixed_version: str,
        vuln_pkg_status: str,
    ):
        # if the vulnerability fixed version is None,
        # that means that no fix has been published
        # yet.
        if vuln_source_fixed_version is None:
            if vuln_pkg_status != "not-vulnerable":
                return True

        return False

    @lru_cache(maxsize=None)
    def _get_installed_source_pkg_version(self, binary_pkg_name: str):
        out, _ = system.subp(
            [
                "dpkg-query",
                "-W",
                "-f=${source:Version}",
                binary_pkg_name,
            ]
        )

        return out

    def is_vulnerability_valid_but_not_fixable(
        self,
        vuln_bin_fix_version: Optional[str],
        bin_pkg_name: str,
        vuln_source_fixed_version: str,
    ):
        """
        This method checks if we can detect that a vulnerability
        affects a binary package but can't be fixed. This
        situation can happen during a package transition.

        For example, suppose we have this entry for pkg1:

        "pkg1": {
          "source_version": {
            "1.0": {
              "bin-pkg1": "1.0",
              "bin-pkg2": "1.1",
            },
            "1.1": {
              "bin-pkg1": "1.2"
            }
          }
        }

        Notice that version 1.1 doesn't produce bin-pkg2 anymore.
        Therefore, if we detect that a vulnerability is fixable
        by version 1.1, we won't find the binary fixable bersion for
        the bin-pkg2 package.

        If we detect that, we will:

        1. Check if versions of the source package associated with the
           binary package is higher than the vulnerability source fix
           version. If it is, we can say that the system is not vulnerable.
        2. If it is not, then the binary package is affected by the issue, but
           we can't say what the user needs to do to fix it.
        """

        if vuln_bin_fix_version is None:
            installed_source_pkg_version = (
                self._get_installed_source_pkg_version(bin_pkg_name)
            )

            if (
                apt.version_compare(
                    installed_source_pkg_version, vuln_source_fixed_version
                )
                > 0
            ):
                return False
            else:
                return True

        return False

    def vulnerability_affects_system(
        self,
        bin_version: str,
        vuln_bin_fix_version: str,
    ):
        return apt.version_compare(vuln_bin_fix_version, bin_version) > 0

    def _list_binary_packages(self, installed_pkgs_by_source: Dict[str, Any]):
        for source_pkg, binary_pkgs in installed_pkgs_by_source.items():
            for (
                binary_pkg_name,
                binary_installed_version,
            ) in sorted(binary_pkgs.items()):
                yield source_pkg, binary_pkg_name, binary_installed_version

    def get_vulnerabilities_for_installed_pkgs(
        self,
        vulnerabilities_data: Dict[str, Any],
        installed_pkgs_by_source: Dict[str, Dict[str, str]],
    ):
        packages = {}  # type: Dict[str, Any]
        vulnerabilities = {}  # type: Dict[str, Any]

        affected_pkgs = vulnerabilities_data.get("packages", {})
        vulns_info = vulnerabilities_data.get("security_issues", {}).get(
            self.vulnerability_type, {}
        )

        for (
            source_pkg,
            bin_pkg_name,
            bin_pkg_version,
        ) in self._list_binary_packages(installed_pkgs_by_source):
            affected_pkg = affected_pkgs.get(source_pkg, {})
            vuln_source_versions = affected_pkg.get("source_versions", {})

            for vuln_name, vuln in sorted(
                self.get_package_vulnerabilities(affected_pkg).items(),
                key=lambda x: x[0],
            ):
                vuln_info = vulns_info.get(vuln_name, "")
                vuln_source_fixed_version = vuln.get("source_fixed_version")
                vuln_pkg_status = vuln.get("status")

                if self.is_vulnerability_not_fixable(
                    vuln_pkg_status=vuln_pkg_status,
                    vuln_source_fixed_version=vuln_source_fixed_version,
                ):
                    self._add_unfixable_vulnerability(
                        packages=packages,
                        bin_pkg_name=bin_pkg_name,
                        bin_pkg_version=bin_pkg_version,
                        vuln_name=vuln_name,
                        vuln_pkg_status=vuln_pkg_status,
                    )
                    self._add_vulnerability_info(
                        vuln_name=vuln_name,
                        vulnerabilities=vulnerabilities,
                        vuln_info=vuln_info,
                        vulns_data=vulnerabilities_data,
                    )
                    continue

                try:
                    pocket = vuln_source_versions[
                        vuln_source_fixed_version
                    ].get("pocket")
                    vuln_bin_fix_version = (
                        vuln_source_versions[vuln_source_fixed_version]
                        .get("binary_packages", {})
                        .get(bin_pkg_name)
                    )
                except KeyError:
                    # There is bug in the data where some sources are
                    # not present. The Security team is already aware
                    # of this issue and they are handling it
                    continue

                if self.is_vulnerability_valid_but_not_fixable(
                    vuln_bin_fix_version,
                    bin_pkg_name,
                    vuln_source_fixed_version,
                ):
                    self._add_unfixable_vulnerability(
                        packages=packages,
                        bin_pkg_name=bin_pkg_name,
                        bin_pkg_version=bin_pkg_version,
                        vuln_name=vuln_name,
                        vuln_pkg_status="unknown",
                    )
                    self._add_vulnerability_info(
                        vuln_name=vuln_name,
                        vulnerabilities=vulnerabilities,
                        vuln_info=vuln_info,
                        vulns_data=vulnerabilities_data,
                    )

                if vuln_bin_fix_version is None:
                    continue

                if self.vulnerability_affects_system(
                    bin_pkg_version,
                    vuln_bin_fix_version,
                ):
                    self._add_fixable_vulnerability(
                        packages=packages,
                        bin_pkg_name=bin_pkg_name,
                        bin_pkg_version=bin_pkg_version,
                        vuln_name=vuln_name,
                        vuln_pkg_status=vuln_pkg_status,
                        vuln_bin_fix_version=vuln_bin_fix_version,
                        vuln_pocket=pocket,
                    )
                    self._add_vulnerability_info(
                        vuln_name=vuln_name,
                        vulnerabilities=vulnerabilities,
                        vuln_info=vuln_info,
                        vulns_data=vulnerabilities_data,
                    )

        return VulnerabilityParserResult(
            vulnerability_data_published_at=vulnerabilities_data.get(
                "published_at"
            ),
            vulnerabilities_info={
                "packages": packages,
                "vulnerabilities": vulnerabilities,
            },
        )


class VulnerabilityResultCache:

    def __init__(self, vulnerability_type: str, series: Optional[str] = None):
        self.series = series or system.get_release_info().series
        self.vulnerability_type = vulnerability_type
        self.dpkg_status_cache = DataObjectFile(
            data_object_cls=VulnerabilityDpkgCacheDate,
            ua_file=UAFile(
                name=VULNERABILITY_DPKG_STATUS_DATE_CACHE,
                directory=VULNERABILITY_CACHE_PATH,
                private=False,
            ),
        )

    def _get_result_cache_path(self):
        return os.path.join(
            VULNERABILITY_CACHE_PATH,
            self.series,
            self.vulnerability_type,
            VULNERABILITY_RESULT_CACHE,
        )

    def save_result_cache(self, vulnerability_data: Dict[str, Any]):
        if util.we_are_currently_root():
            latest_dpkg_status_time = apt.get_dpkg_status_time() or 0
            self.dpkg_status_cache.write(
                VulnerabilityDpkgCacheDate(
                    dpkg_status_date=latest_dpkg_status_time
                )
            )
            system.write_file(
                self._get_result_cache_path(),
                json.dumps(vulnerability_data),
            )

    def _has_apt_state_changed(self):
        latest_dpkg_status_time = apt.get_dpkg_status_time() or 0
        dpkg_status_cache_obj = self.dpkg_status_cache.read()
        if not dpkg_status_cache_obj:
            return True

        return latest_dpkg_status_time > dpkg_status_cache_obj.dpkg_status_date

    def _cache_result_exists(self):
        return os.path.exists(self._get_result_cache_path())

    def _is_cache_result_valid(self):
        if not self._cache_result_exists():
            return False

        if self._has_apt_state_changed():
            return False

        return True

    def is_cache_valid(self):
        return self._is_cache_result_valid()

    def get_result_cache(self):
        return json.loads(system.load_file(self._get_result_cache_path()))


def get_vulnerabilities(
    parser: VulnerabilityParser,
    cfg: UAConfig,
    series: Optional[str],
):
    vulnerabilities_data = VulnerabilityData(
        cfg=cfg,
        series=series,
    )
    vulnerabilities_result = VulnerabilityResultCache(
        series=series,
        vulnerability_type=parser.vulnerability_type,
    )

    vulnerabilities_json_data = vulnerabilities_data.get()

    if not vulnerabilities_data.refreshed:
        if vulnerabilities_result.is_cache_valid():
            return VulnerabilityParserResult(
                vulnerability_data_published_at=vulnerabilities_data.get_published_date(),  # noqa
                vulnerabilities_info=vulnerabilities_result.get_result_cache(),  # noqa
            )

    installed_pkgs_by_source = query_installed_source_pkg_versions()

    vulnerabilities_parser_result = (
        parser.get_vulnerabilities_for_installed_pkgs(
            vulnerabilities_data=vulnerabilities_json_data,
            installed_pkgs_by_source=installed_pkgs_by_source,
        )
    )

    vulnerabilities_result.save_result_cache(
        vulnerabilities_parser_result.vulnerabilities_info
    )

    return vulnerabilities_parser_result



from typing import List

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import (
    DataObject,
    Field,
    IntDataValue,
    StringDataValue,
    data_list,
)
from uaclient.security_status import (
    create_updates_list,
    filter_updates,
    get_installed_packages_by_origin,
    get_ua_info,
)


class UpdateSummary(DataObject):
    fields = [
        Field(
            "num_updates", IntDataValue, doc="Total count of available updates"
        ),
        Field(
            "num_esm_apps_updates",
            IntDataValue,
            doc="Count of available updates from ``esm-apps``",
        ),
        Field(
            "num_esm_infra_updates",
            IntDataValue,
            doc="Count of available updates from ``esm-infra``",
        ),
        Field(
            "num_standard_security_updates",
            IntDataValue,
            doc="Count of available updates from the ``-security`` pocket",
        ),
        Field(
            "num_standard_updates",
            IntDataValue,
            doc="Count of available updates from the ``-updates`` pocket",
        ),
    ]

    def __init__(
        self,
        num_updates: int,
        num_esm_apps_updates: int,
        num_esm_infra_updates: int,
        num_standard_security_updates: int,
        num_standard_updates: int,
    ):
        self.num_updates = num_updates
        self.num_esm_apps_updates = num_esm_apps_updates
        self.num_esm_infra_updates = num_esm_infra_updates
        self.num_standard_security_updates = num_standard_security_updates
        self.num_standard_updates = num_standard_updates


class UpdateInfo(DataObject):
    fields = [
        Field(
            "download_size",
            IntDataValue,
            doc="Download size for the update in bytes",
        ),
        Field(
            "origin",
            StringDataValue,
            doc="Where the update is downloaded from",
        ),
        Field(
            "package", StringDataValue, doc="Name of the package to be updated"
        ),
        Field(
            "provided_by",
            StringDataValue,
            doc="Service which provides the update",
        ),
        Field(
            "status",
            StringDataValue,
            doc="Whether this update is ready for download or not",
        ),
        Field("version", StringDataValue, doc="Version of the update"),
    ]

    def __init__(
        self,
        download_size: int,
        origin: str,
        package: str,
        provided_by: str,
        status: str,
        version: str,
    ):
        self.download_size = download_size
        self.origin = origin
        self.package = package
        self.provided_by = provided_by
        self.status = status
        self.version = version


class PackageUpdatesResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "summary", UpdateSummary, doc="Summary of all available updates"
        ),
        Field(
            "updates",
            data_list(UpdateInfo),
            doc="Detailed list of all available updates",
        ),
    ]

    def __init__(self, summary: UpdateSummary, updates: List[UpdateInfo]):
        self.summary = summary
        self.updates = updates


def updates() -> PackageUpdatesResult:
    return _updates(UAConfig())


def _updates(cfg: UAConfig) -> PackageUpdatesResult:
    """
    This endpoint shows available updates for packages in a system, categorized
    by where they can be obtained.
    """
    ua_info = get_ua_info(cfg)
    packages = get_installed_packages_by_origin()
    upgradable_versions = filter_updates(packages["all"])
    update_list = create_updates_list(upgradable_versions, ua_info)

    num_esm_apps_updates = len(upgradable_versions["esm-apps"])
    num_esm_infra_updates = len(upgradable_versions["esm-infra"])
    num_standard_security_updates = len(
        upgradable_versions["standard-security"]
    )
    num_standard_updates = len(upgradable_versions["standard-updates"])

    summary = UpdateSummary(
        num_updates=num_esm_apps_updates
        + num_esm_infra_updates
        + num_standard_security_updates
        + num_standard_updates,
        num_esm_apps_updates=num_esm_apps_updates,
        num_esm_infra_updates=num_esm_infra_updates,
        num_standard_security_updates=num_standard_security_updates,
        num_standard_updates=num_standard_updates,
    )
    updates = [
        UpdateInfo(
            download_size=update["download_size"],
            origin=update["origin"],
            package=update["package"],
            provided_by=update["service_name"],
            status=update["status"],
            version=update["version"],
        )
        for update in update_list
    ]

    return PackageUpdatesResult(summary=summary, updates=updates)


endpoint = APIEndpoint(
    version="v1",
    name="PackageUpdates",
    fn=_updates,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.packages.updates.v1 import updates

result = updates()
""",  # noqa: E501
    "result_class": PackageUpdatesResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.packages.updates.v1",
    "example_json": """
{
    "summary": {
        "num_updates": 1,
        "num_esm_apps_updates": 2,
        "num_esm_infra_updates": 3,
        "num_standard_security_updates": 4,
        "num_standard_updates": 5,
    },
    "updates": [
        {
            "download_size": 6,
            "origin": "<some site>",
            "package": "<package name>",
            "provided_by": "<service name>",
            "status": "<update status>",
            "version": "<updated version>",
        },
    ]
}
""",
}



"""Tests related to uaclient.snap module."""

import mock
import pytest

from uaclient import exceptions, messages
from uaclient.snap import (
    SnapPackage,
    configure_snap_proxy,
    get_config_option_value,
    get_installed_snaps,
    get_snap_info,
    is_snapd_installed_as_a_snap,
    unconfigure_snap_proxy,
)


class TestIsSnapdInstalledAsASnap:
    @pytest.mark.parametrize(
        ["installed_snaps", "expected"],
        [
            ([], False),
            (
                [
                    SnapPackage(
                        "one", "123", "456", "oldest/unstable", "someone"
                    )
                ],
                False,
            ),
            (
                [
                    SnapPackage(
                        "snapd", "123", "456", "oldest/unstable", "someone"
                    )
                ],
                True,
            ),
            (
                [
                    SnapPackage(
                        "one", "123", "456", "oldest/unstable", "someone"
                    ),
                    SnapPackage(
                        "snapd", "123", "456", "oldest/unstable", "someone"
                    ),
                ],
                True,
            ),
        ],
    )
    @mock.patch("uaclient.snap.get_installed_snaps")
    def test_is_snapd_installed_as_a_snap(
        self, m_get_installed_snaps, installed_snaps, expected
    ):
        m_get_installed_snaps.return_value = installed_snaps
        assert expected == is_snapd_installed_as_a_snap()


class TestConfigureSnapProxy:
    @pytest.mark.parametrize("http_proxy", ("http_proxy", "", None))
    @pytest.mark.parametrize("https_proxy", ("https_proxy", "", None))
    @pytest.mark.parametrize("snapd_installed", (True, False))
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.snap.is_snapd_installed")
    def test_configure_snap_proxy(
        self,
        m_is_snapd_installed,
        m_subp,
        http_proxy,
        https_proxy,
        snapd_installed,
        capsys,
    ):
        retry_sleeps = mock.MagicMock()
        m_is_snapd_installed.return_value = snapd_installed
        configure_snap_proxy(http_proxy, https_proxy, retry_sleeps)
        expected_calls = []
        if snapd_installed:
            if http_proxy:
                expected_calls.append(
                    mock.call(
                        [
                            "snap",
                            "set",
                            "system",
                            "proxy.http={}".format(http_proxy),
                        ],
                        retry_sleeps=retry_sleeps,
                    )
                )

            if https_proxy:
                expected_calls.append(
                    mock.call(
                        [
                            "snap",
                            "set",
                            "system",
                            "proxy.https={}".format(https_proxy),
                        ],
                        retry_sleeps=retry_sleeps,
                    )
                )

        assert m_subp.call_args_list == expected_calls
        assert 1 == m_is_snapd_installed.call_count

        out, _ = capsys.readouterr()
        if snapd_installed and (http_proxy or https_proxy):
            assert out.strip() == messages.SETTING_SERVICE_PROXY.format(
                service="snap"
            )

    @pytest.mark.parametrize(
        "key, subp_side_effect, expected_ret",
        [
            (
                "proxy.http",
                exceptions.ProcessExecutionError("doesn't matter"),
                None,
            ),
            ("proxy.https", ("value", ""), "value"),
        ],
    )
    @mock.patch("uaclient.system.subp")
    def test_get_config_option_value(
        self, m_util_subp, key, subp_side_effect, expected_ret
    ):
        m_util_subp.side_effect = [subp_side_effect]
        ret = get_config_option_value(key)
        assert ret == expected_ret
        assert [
            mock.call(["snap", "get", "system", key])
        ] == m_util_subp.call_args_list


class TestUnconfigureSnapProxy:
    @pytest.mark.parametrize("protocol_type", ("http", "https"))
    @pytest.mark.parametrize("retry_sleeps", (None, [1], []))
    @pytest.mark.parametrize("snapd_installed", (True, False))
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.snap.is_snapd_installed")
    def test_unconfigure_snap_proxy(
        self,
        m_snapd_installed,
        m_subp,
        protocol_type,
        retry_sleeps,
        snapd_installed,
    ):
        m_snapd_installed.return_value = snapd_installed
        subp_calls = []
        if snapd_installed:
            subp_calls = [
                mock.call(
                    ["snap", "unset", "system", "proxy." + protocol_type],
                    retry_sleeps=retry_sleeps,
                )
            ]

        kwargs = {"protocol_type": protocol_type}
        if retry_sleeps is not None:
            kwargs["retry_sleeps"] = retry_sleeps
        assert None is unconfigure_snap_proxy(**kwargs)
        assert 1 == m_snapd_installed.call_count
        assert subp_calls == m_subp.call_args_list


class TestSnapPackagesInstalled:
    @mock.patch("uaclient.snap.get_snap_info")
    @mock.patch("uaclient.snap.system.subp")
    def test_snap_packages_installed(self, m_sys_subp, m_get_snap_info):
        m_sys_subp.return_value = (
            "Name  Version Rev Tracking Publisher Notes\n"
            "helloworld 6.0.16 126 latest/stable dev1 -\n"
            "bare 1.0 5 latest/stable canonical** base\n"
            "canonical-livepatch 10.2.3 146 latest/stable canonical** -\n"
        ), ""
        expected_snaps = [
            SnapPackage(
                "helloworld",
                "6.0.16",
                "126",
                "latest/stable",
                "dev1",
            ),
            SnapPackage(
                "bare",
                "1.0",
                "5",
                "latest/stable",
                "canonical**",
            ),
            SnapPackage(
                "canonical-livepatch",
                "10.2.3",
                "146",
                "latest/stable",
                "canonical**",
            ),
        ]
        m_get_snap_info.side_effect = expected_snaps
        snaps = get_installed_snaps()
        assert snaps[0].name == expected_snaps[0].name
        assert snaps[0].revision == expected_snaps[0].revision
        assert snaps[1].name == expected_snaps[1].name
        assert snaps[1].publisher == expected_snaps[1].publisher
        assert snaps[2].channel == expected_snaps[2].channel


class TestGetSnapInfo:
    @pytest.mark.parametrize(
        "status_code,response_data,expected_exception",
        (
            (
                404,
                {"result": {"kind": "snap-not-found"}},
                exceptions.SnapNotInstalledError,
            ),
            (
                404,
                {"result": {"kind": ""}},
                exceptions.UnexpectedSnapdAPIError,
            ),
            (
                500,
                {"result": {"message": "error"}},
                exceptions.UnexpectedSnapdAPIError,
            ),
            (
                401,
                {},
                exceptions.UnexpectedSnapdAPIError,
            ),
        ),
    )
    @mock.patch("uaclient.snap.http.unix_socket_request")
    def test_get_snap_info_unexpected_api_error(
        self,
        m_unix_socket_request,
        status_code,
        response_data,
        expected_exception,
    ):
        m_unix_socket_request.return_value = mock.MagicMock(
            code=status_code, json_dict=response_data
        )

        with pytest.raises(expected_exception):
            get_snap_info(snap="test")

        assert 1 == m_unix_socket_request.call_count

    @mock.patch("uaclient.snap.http.unix_socket_request")
    def test_get_snap_info_connection_refused(
        self,
        m_unix_socket_request,
    ):
        m_unix_socket_request.side_effect = ConnectionRefusedError()

        with pytest.raises(exceptions.SnapdAPIConnectionRefused):
            get_snap_info(snap="test")

        assert 1 == m_unix_socket_request.call_count

    @mock.patch("uaclient.snap.http.unix_socket_request")
    def test_get_snap_info(
        self,
        m_unix_socket_request,
    ):
        m_unix_socket_request.return_value = mock.MagicMock(
            code=200,
            json_dict={
                "result": {
                    "channel": "stable",
                    "revision": "120",
                    "name": "test",
                    "publisher": {"username": "canonical"},
                }
            },
        )

        expected_result = SnapPackage(
            name="test",
            version="",
            revision="120",
            channel="stable",
            publisher="canonical",
        )
        assert expected_result == get_snap_info(snap="test")

        assert 1 == m_unix_socket_request.call_count



from uaclient import system
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.attach.auto.full_auto_attach.v1 import (
    FullAutoAttachOptions,
)
from uaclient.config import UAConfig
from uaclient.daemon import retry_auto_attach
from uaclient.data_types import DataObject
from uaclient.files import state_files

ConfigureRetryServiceOptions = FullAutoAttachOptions


class ConfigureRetryServiceResult(DataObject, AdditionalInfo):
    pass


def configure_retry_service(
    options: ConfigureRetryServiceOptions,
) -> ConfigureRetryServiceResult:
    return _configure_retry_service(options, UAConfig())


def _configure_retry_service(
    options: ConfigureRetryServiceOptions, cfg: UAConfig
) -> ConfigureRetryServiceResult:
    """
    This endpoint configures options for the retry auto-attach functionality,
    and creates files that will activate the retry auto-attach functionality
    if ``ubuntu-advantage.service`` runs.

    Note that this does not start ``ubuntu-advantage.service``. This makes it
    useful for calling during the boot process
    ``Before: ubuntu-advantage.service`` so that when
    ``ubuntu-advantage.service`` starts, its ``ConditionPathExists`` check
    passes and activates the retry auto-attach function.

    If you call this function outside of the boot process and would like the
    retry auto-attach functionality to actually start, you'll need to call
    something like ``systemctl start ubuntu-advantage.service``.
    """
    state_files.retry_auto_attach_options_file.write(
        state_files.RetryAutoAttachOptions(
            enable=options.enable, enable_beta=options.enable_beta
        )
    )
    system.create_file(retry_auto_attach.FLAG_FILE_PATH)
    return ConfigureRetryServiceResult()


endpoint = APIEndpoint(
    version="v1",
    name="ConfigureRetryService",
    fn=_configure_retry_service,
    options_cls=ConfigureRetryServiceOptions,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "extra_args_content": """
.. note::

    If none of the lists are set, the services will be enabled based on the
    contract definitions.
""",
    "example_python": """
from uaclient.api.u.pro.attach.auto.configure_retry_service.v1 import configure_retry_service, ConfigureRetryServiceOptions

options = ConfigureRetryServiceOptions(enable=["<service1>", "<service2>"], enable_beta=["<beta_service3>"])
result = configure_retry_service(options)
""",  # noqa: E501
    "result_class": ConfigureRetryServiceResult,
    "exceptions": [],
    "example_cli": 'pro api u.pro.attach.auto.configure_retry_service.v1 --data {"enable": ["esm-infra", "esm-apps"]}',  # noqa: E501
    "example_json": """
{}
""",
}



from uaclient import config, contract, system, util
from uaclient.api import exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.data_types import (
    DataObject,
    DatetimeDataValue,
    Field,
    StringDataValue,
)
from uaclient.files import machine_token


class GetGuestTokenResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "guest_token",
            StringDataValue,
            doc="The guest token",
        ),
        Field(
            "id",
            StringDataValue,
            doc="The ID of the guest token",
        ),
        Field(
            "expires",
            DatetimeDataValue,
            doc="The expiration time of the guest token",
        ),
    ]

    def __init__(
        self,
        guest_token: str,
        id: str,
        expires: str,
    ):
        self.guest_token = guest_token
        self.id = id
        self.expires = expires


def get_guest_token() -> GetGuestTokenResult:
    return _get_guest_token(config.UAConfig())


def _get_guest_token(
    cfg: config.UAConfig,
) -> GetGuestTokenResult:
    """
    This endpoint fetches a guest token from the backend.
    """
    if not util.we_are_currently_root():
        raise exceptions.NonRootUserError()
    if not _is_attached(cfg).is_attached:
        raise exceptions.UnattachedError()
    machine_token_file = machine_token.get_machine_token_file(cfg)
    machine_token_dict = machine_token_file.machine_token
    contract_id = machine_token_file.contract_id
    if machine_token_dict is None or contract_id is None:
        # mypy doesn't know that these will never be None if we're attached
        # and root
        raise exceptions.UnattachedError()
    machine_token_str = machine_token_dict["machineToken"]
    machine_id = system.get_machine_id(cfg)
    contract_client = contract.UAContractClient(cfg)
    guest_token = contract_client.get_guest_token(
        machine_token=machine_token_str,
        contract_id=contract_id,
        machine_id=machine_id,
    )
    return GetGuestTokenResult(
        guest_token=guest_token["guestToken"],
        id=guest_token["id"],
        expires=guest_token["expires"],
    )


endpoint = APIEndpoint(
    version="v1",
    name="GetGuestToken",
    fn=_get_guest_token,
    options_cls=None,
)

_doc = {
    "introduced_in": "35",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.attach.guest_token.get_guest_token.v1 import get_guest_token
get_guest_token()
""",  # noqa: E501
    "result_class": GetGuestTokenResult,
    "exceptions": [
        (
            exceptions.ConnectivityError,
            (
                "Raised if it is not possible to connect to the contracts"
                " service."
            ),
        ),
        (
            exceptions.ContractAPIError,
            (
                "Raised if there is an unexpected error in the contracts"
                " service interaction."
            ),
        ),
        (
            exceptions.NonRootUserError,
            ("Raised if a non-root user executes this endpoint."),
        ),
        (exceptions.UnattachedError, "Raised if the machine is not attached"),
        (
            exceptions.FeatureNotSupportedOldTokenError,
            ("Raised if the machine needs to be re-attached first."),
        ),
    ],
    "example_cli": "pro api u.pro.attach.guest_token.get_guest_token.v1",
    "example_json": """
{
    "guest_token": "...",
    "id": "...",
    "expires": "..."
}
""",
}



import logging

from uaclient import config, exceptions, http, log, secret_manager, util
from uaclient.clouds import AutoAttachInstance

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

LXD_INSTANCE_API_SOCKET_PATH = "/dev/lxd/sock"
LXD_INSTANCE_API_ENDPOINT_UBUNTU_PRO = "/1.0/ubuntu-pro"
LXD_INSTANCE_API_ENDPOINT_UBUNTU_PRO_GUEST_TOKEN = "/1.0/ubuntu-pro/token"


class LXDAutoAttachInstance(AutoAttachInstance):
    @property
    def is_viable(self) -> bool:
        return True

    def should_poll_for_pro_license(self) -> bool:
        """Yes, but only once - is_pro_license_present doesn't
        support wait_for_change"""
        return True

    def is_pro_license_present(self, *, wait_for_change: bool) -> bool:
        if wait_for_change:
            # Unsupported
            raise exceptions.CancelProLicensePolling()

        resp = http.unix_socket_request(
            LXD_INSTANCE_API_SOCKET_PATH,
            "GET",
            LXD_INSTANCE_API_ENDPOINT_UBUNTU_PRO,
        )
        if resp.code != 200:
            LOG.error(
                "LXD instance API returned error for ubuntu-pro query",
                extra=log.extra(code=resp.code, body=resp.body),
            )
            return False

        # returning True will cause auto-attach on launch, so only "on" counts
        if resp.json_dict.get("guest_attach", "off") != "on":
            LOG.info(
                "guest_attach config is not turned on. The machine will not try to auto-attach."  # noqa
            )
            return False

        return True

    def acquire_pro_token(self, cfg: config.UAConfig) -> str:
        """
        Cloud-specific implementation of acquiring the pro token using whatever
        method suits the platform
        """
        resp = http.unix_socket_request(
            LXD_INSTANCE_API_SOCKET_PATH,
            "POST",
            LXD_INSTANCE_API_ENDPOINT_UBUNTU_PRO_GUEST_TOKEN,
        )
        if resp.code == 404:
            raise exceptions.LXDAutoAttachNotAvailable()
        elif resp.code == 403:
            raise exceptions.LXDAutoAttachNotAllowed()
        elif resp.code != 200:
            raise exceptions.ExternalAPIError(
                code=resp.code,
                url="unix://{}{}".format(
                    LXD_INSTANCE_API_SOCKET_PATH,
                    LXD_INSTANCE_API_ENDPOINT_UBUNTU_PRO_GUEST_TOKEN,
                ),
                body=resp.body,
            )
        guest_token = resp.json_dict.get("guest_token", "")
        secret_manager.secrets.add_secret(guest_token)
        return guest_token
